/*
    screen_playing.c - User Interface
     Copyright (C) 2009 by H. Raap

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "screen.h"
#include "window.h"
#include "keyboard.h"
#include "lcd.h"
#include "model.h"
#include "mpd.h"


/* Number of entries in the window list */
#define WL_SIZE 9

/* The window list */
static struct Window win[WL_SIZE];

#define WIN_TXT_SIZE 80
/* Text buffer for the windows */
static char win_txt[WL_SIZE][WIN_TXT_SIZE];

/* Some definitions to refer to specific windows by name */

/* Window to show the current title */
#define title_win (win[0])

/* Window to show the current artist */
#define artist_win (win[1])

/* Window to show running time and total time */
#define time_win (win[2])

/* Window to show a progress bar for running time */
#define time_prog_win (win[3])

/* Window to show volume ramp */
#define vol_ramp_win (win[4])

/* Window to show the current volume in precent */
#define volume_win (win[5])

/* Window to show speaker on/off */
#define speaker_win (win[6])

/* Window to show play/stop/pause status */
#define state_win (win[7])

#define version_win (win[8])

/* Window to show the current album TODO maybe later */

	
/* These variables are set if we detect that the user wants something */
static enum USER_CMD user_cmd;
static int user_arg;


/* Initialize the playing screen (showing current title and status) 
	We clear the whole screen.
*/
void 
playing_screen_init(void) {
	user_cmd = NO_CMD;
	
	win_init(&title_win, 0, 0, 50, 128, 1, win_txt[0]);
	title_win.font = MEDIUMFONT;	
	title_win.flags |= WINFLG_CENTER;
	
	win_init(&artist_win, 49, 0, 30, 128, 1, win_txt[1]);
	artist_win.font = MEDIUMFONT;	
	artist_win.flags |= WINFLG_CENTER;
	
	win_init(&time_win, 84, 22, 10, 128-22, 0, win_txt[2]);
	time_win.font = SMALLFONT;

	/* The state window, on the same line as the time progress */
	win_init(&state_win, 89, 1, 16, 18, 0, win_txt[3]);

	state_win.font = BIGFONT;
	win_new_text(&state_win, "\xB3"); 	// play icon == b3
	
	/* This is our progressbar */
	win_init(&time_prog_win, 95, 22, 8, 102, 1, win_txt[4]);
	time_prog_win.font = SMALLFONT;
	time_prog_win.flags = WINFLG_BAR;
	time_prog_win.fg_color = DARK_GREY;
	time_prog_win.cur_char = 0;		/* current value */
	
#define VOL_ROW 114	
	/* The speaker indicator is on the same line as volume text */
	win_init(&speaker_win, VOL_ROW, 0, 18, 30, 0, win_txt[5]);
	speaker_win.font = BIGFONT;
	win_new_text(&speaker_win, "\xB9\xB2");	// speaker icon on
	
	/* Our indicator for the volume (triangle progress bar) */
	win_init(&vol_ramp_win, VOL_ROW, 31, 16, 50, 0, win_txt[6]);
	vol_ramp_win.flags = WINFLG_RAMP;
	vol_ramp_win.fg_color = DARK_GREY;
	vol_ramp_win.cur_char = 100/2;		/* We initialize the volume with 100 % */

	/* The text portion of the volume follows on the same line as the ramp */
	win_init(&volume_win, VOL_ROW, 88, 16, 128-88, 0, win_txt[7]);
	win_new_text(&volume_win, "100 %");

	win_init(&version_win, 150, 0, 10, 128, 0, win_txt[8]); 
	version_win.font = SMALLFONT;	
	version_win.flags |= WINFLG_CENTER;
	version_win.fg_color = WHITE;
	version_win.bg_color = BLACK;
	win_new_text(&version_win, "Version " VERSION);

}

static void
playing_screen_visible(int v){
	int i;
	for (i=0; i < WL_SIZE; i++)
		if (v) 
			win[i].flags |= WINFLG_VISIBLE;
		else
			win[i].flags &= ~WINFLG_VISIBLE;
};


/* redraw all the windows 
	assumes the screen has just been cleared
*/
static void
playing_screen_redraw(){
	int i;
	
	for (i=0; i < WL_SIZE; i++)
		win_redraw(&win[i]);
};


void 
playing_screen_enter(){
	lcd_fill(0x00);
	playing_screen_visible(1);
	version_win.flags &= ~WINFLG_VISIBLE;
	playing_screen_redraw();	
};

void
playing_screen_exit(){
	playing_screen_visible(0);
};

void
view_time_changed(int te, int tt){
	char newstr[50];
	int newval;
	
	sec2hms(newstr, te);
	strn_cat(newstr, " / ", 3);	
	sec2hms(newstr+8, tt);
	win_new_text(&time_win, newstr);
		
	newval = (te * 100) / tt;
	
	if ( (win->flags & WINFLG_VISIBLE))
		draw_progressbar(time_prog_win.start_row+1, time_prog_win.start_col, time_prog_win.height-2, 
							time_prog_win.fg_color, time_prog_win.bg_color, time_prog_win.cur_char, newval);
	time_prog_win.cur_char = newval;
};


void 
view_title_changed(char *s){
	win_new_text(&title_win, s);
};

void 
view_artist_changed(char *s){
	win_new_text(&artist_win, s);
};


/* The volume has changed. Show new value. */
void
view_volume_changed(int vol){
	char newstr[250];
	char number[20];
	char *numstr;
		
	if (vol > 0)
		win_new_text(&speaker_win, "\xB9\xB2");	// speaker icon on
	else
		win_new_text(&speaker_win, "\xB9  ");	//speaker off	
	
	strn_cpy (newstr, "", 1);
	if (vol < 0)
		strn_cat(newstr, "?? \n", 12);
	else {
		numstr = get_digits(vol, number, 0); 
		strn_cat(newstr, numstr, 12);
		strn_cat(newstr, " %\n", 2);
	};
	win_new_text(&volume_win, newstr);	

	if (vol >= 0){
		if ((win->flags & WINFLG_VISIBLE)) 
			draw_trianglebar( vol_ramp_win.start_row+1, vol_ramp_win.start_col+1, vol_ramp_win.width-2,
						  vol_ramp_win.height-2, vol_ramp_win.fg_color, vol_ramp_win.bg_color, vol_ramp_win.cur_char, vol/2 );
			vol_ramp_win.cur_char = vol/2;
	};
};

void
view_state_changed(enum PLAYSTATE state){
	if (state == PLAY)
		win_new_text(&state_win, "\xB3"); 	// play icon == b3
	else
		if (state == PAUSE)
			win_new_text(&state_win, "\xBA"); 	// pause icon == ba
	else
		if (state == STOP)
			win_new_text(&state_win, "\xA0"); 	// pause icon == a0
	
};


/* Returns user command if there is any, else returns NO_CMD */
enum USER_CMD 
playing_user_request(void){
	enum USER_CMD res = user_cmd;
	user_cmd = NO_CMD;
	return (res);
};

/* Sometimes we need an argument for the user cmd, this is returned by this function */
int 
playing_arg(void){
	return user_arg;
};


void 
playing_keypress(int cur_key){
	switch (cur_key) {

		case KEY_Down:
		case KEY_Yellow:
			user_cmd = PAUSE_TOGGLE;
			break;
			
		case KEY_Up:
		case KEY_Green:
			user_cmd = PLAY_CMD;
			break;
			
		case KEY_Left:
			user_cmd = REWIND_CMD;
			break;
			
		case KEY_Right:
			user_cmd = FORWARD_CMD;
			break;
			
		case KEY_Exit:
			switch_screen(PLAYING_SCREEN, TRACKLIST_SCREEN);
			break;
			
		case KEY_Red:
			user_cmd = STOP_CMD;
			break;

		case KEY_Vplus:	
			/* Inform the controller that the user wants a volume increase */
				user_cmd = VOLUME_UP;
			break;
			
		case KEY_Vminus:
			/* Inform the controller that the user wants a volume decrease */
				user_cmd = VOLUME_DOWN;
			break;
		
		case KEY_Mute:	
			/* Inform the controller that the user wants a volume increase */
				user_cmd = MUTE_CMD;
			break;
			
			/* Inform the controller that the user wants the next song */
		case KEY_Pplus:
			user_cmd = NEXT_SONG;
			break;
			
			/* Inform the controller that the user wants the previous song */
		case KEY_Pminus:
			user_cmd = PREV_SONG;
			break;
			
			/* Show version info */
		case KEY_Info:
			if ( (version_win.flags & WINFLG_VISIBLE)){
				version_win.bg_color = WHITE;
				win_clear(&version_win, 1); 
				version_win.flags &= ~WINFLG_VISIBLE;
				version_win.bg_color = BLACK;
			} else {
				version_win.flags |= WINFLG_VISIBLE;
				win_redraw(&version_win);
			}; 
			break;
			
		case KEY_VTX1:
			fDebug ^= 1;	
			break;
			
		default:
			break;
	};
};	
