/*
    screen.c - User Interface
	Copyright (C) 2009 by H. Raap

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "screen.h"
#include "window.h"
#include "keyboard.h"
#include "lcd.h"
#include "model.h"
#include "mpd.h"

/* ------------------------------------- Track List Menu ------------------------------------------ */
/* The row where the window list starts */
#define WL_START_ROW 0

/* The width of each window in pixels */
#define WL_WIDTH 128

/* The height of a small window */
#define WL_SMALL_HEIGHT 10

/* The height of a non-highlighted window in the window list */
#define WL_NORMAL_HEIGHT 13

/* The height of a highlighted window in the window list */
#define WL_HIGH_HEIGHT 18

/* Number of entries in the window list 
	We have 160 pixels, so 1 small, 10 normal windows and 1 big window fit 
*/
#define WL_SIZE 12

/* The window list */
static struct Window win[WL_SIZE];

#define WIN_TXT_SIZE 32
/* Text buffer for the windows */
static char win_txt[WL_SIZE][WIN_TXT_SIZE];

/* This window is currently selected by the user */
static int sel_win;

// TODO just a hack here with global variables
extern int pl_num;
extern char pl_name[PL_LIM][PL_NAME_LENGTH];
		
/* These variables are set if we detect that the user wants something */
static enum USER_CMD user_cmd;
static int user_arg;



/* redraw all the windows 
	assumes the screen has just been cleared
*/
static void
playlist_screen_redraw(){
	int i;

	for (i=0; i < WL_SIZE; i++)
		win_redraw(&win[i]);
};

static void
playlist_screen_visible(int v){
	int i;
	for (i=0; i < WL_SIZE; i++)
		if (v) 
			win[i].flags |= WINFLG_VISIBLE;
		else
			win[i].flags &= ~WINFLG_VISIBLE;
};

/* We are called when the playlist names have potentially changed (i.e they have been read in from MPD) */
void
view_playlist_changed(){
	char *name;
	int i;
	for (i=1; i < WL_SIZE; i++){
		name = pl_get_name(i - 1);
		if (name == NULL)			// There are no more playlist names. NOTE We do not use the info here, but should!
			name = "";
		win_new_text(&win[i], name);
	};
};


static void
sel_win_up(){
	if (sel_win > 1){

		win[sel_win].font = MEDIUMFONT;
		win[sel_win].height = WL_NORMAL_HEIGHT;
		win[sel_win].start_row = win[sel_win-1].start_row + WL_HIGH_HEIGHT;
		win[sel_win].border = 0;
		sel_win--;						
		win[sel_win].font = BIGFONT;
		win[sel_win].height = WL_HIGH_HEIGHT;
		win[sel_win].border = 1;
				
		win_redraw(&win[sel_win]);
		win_redraw(&win[sel_win+1]);
	};
};

static void
sel_win_down(){
	if (sel_win < WL_SIZE - 1){
		win[sel_win].font = MEDIUMFONT;
		win[sel_win].height = WL_NORMAL_HEIGHT;
		win[sel_win].border = 0;
		sel_win++;		
		win[sel_win].font = BIGFONT;
		win[sel_win].height = WL_HIGH_HEIGHT;
		win[sel_win].border = 1;
		win[sel_win].start_row = win[sel_win-1].start_row + WL_NORMAL_HEIGHT;

		win_redraw(&win[sel_win]);
		win_redraw(&win[sel_win-1]);
	};
};

void
playlist_screen_init(){
	int i;
	int cur_start_row;
	
	user_cmd = NO_CMD;
	cur_start_row = WL_START_ROW;	
	
	/* The first window is a help window */
	win_init(&win[0], cur_start_row, 0, WL_SMALL_HEIGHT, 128,0, win_txt[0]);
	win[0].font = SMALLFONT;
	win[0].fg_color = WHITE;
	win[0].bg_color = BLACK;
	win[0].flags |= WINFLG_CENTER;
	cur_start_row += WL_SMALL_HEIGHT;
	
	sel_win = 1;	
	
	/* The next window is bigger than usual */ 
	
	/* The remaining windows are playlist windows */
	for (i=1; i < WL_SIZE; i++){
		if (i==sel_win){
			/* The selected window is bigger than usual */ 	
			win_init(&win[1], cur_start_row, 0, WL_HIGH_HEIGHT, 128, 1, win_txt[i]);
			win[i].font = BIGFONT;
			cur_start_row += WL_HIGH_HEIGHT;	
		} else {
			win_init(&win[i], cur_start_row, 0, WL_NORMAL_HEIGHT, 128, 0, win_txt[i]);
			cur_start_row += WL_NORMAL_HEIGHT;
		};	
		/* Every second window has a different background color */
		if (0 == (i & 1)) win[i].bg_color = LIGHT_GREY;
	};
	
	win_new_text(&win[0], "Choose a playlist");
};	

void
playlist_screen_enter(){
	
	/* We always want to start with the first playlist name, because the user might have got lost previously. */
	while ( sel_win < 1 )		// should not happen
		sel_win_down();
		
	while ( sel_win  > 1)
		sel_win_up();

	lcd_fill(0x00);	
		
	/* A lot of things might have changed since we were called last.
		So better reread the information from mpd 
	*/
	view_playlist_changed();
	
	playlist_screen_visible(1);
	playlist_screen_redraw();	
};

void
playlist_screen_exit(){
	playlist_screen_visible(0);	
};


/* Returns user command if there is any, else returns NO_CMD */
enum USER_CMD 
playlist_user_request(void){
	enum USER_CMD res = user_cmd;
	user_cmd = NO_CMD;
	return (res);
};

/* Sometimes we need an argument for the user cmd, this is returned by this function */
int 
playlist_arg(void){
	return user_arg;
};


void
playlist_keypress(int cur_key){
	switch (cur_key) {

		case KEY_OK:
			user_cmd = LOAD_CMD;
			user_arg = sel_win - 1;
			switch_screen(PLAYLIST_SCREEN, PLAYING_SCREEN);
			break;
		
		case KEY_Exit:
			switch_screen(PLAYLIST_SCREEN, PLAYING_SCREEN);
			break;
			
		case KEY_Up:
			sel_win_up();
			break;
			
		case KEY_Down:
			sel_win_down();
			break;
			
		case KEY_Info:
			break;		// We prevent a punch through 
			
			/* Punch Through the unknown keys */	
		default:
			playing_keypress(cur_key);
			break;
	};
};
