/*
    screen_tracklist.c - User Interface
	Copyright (C) 2009 by H. Raap

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "screen.h"
#include "window.h"
#include "keyboard.h"
#include "lcd.h"
#include "model.h"
#include "mpd.h"


// TODO just a hack here with global variables

/* We have an externally maintained list (by model.c) with the track names of 20 or 30 tracks.
	The number of tracks currently in this list is returned by a call to num_tracks().
	We can tell mpd that we need some track information by calling get_track_info(track_no).
	Track no. is relative to the current playlist.
	The information might arrive asynchronously.
	We are informed about new information when mpd calls our view_tracklist_changed() routine.
	We only have a limited number of windows to show the track list.
	So we only show a small part of the current track list.
	A good guess is to show the last two tracks, the current track and then the next tracks as far as our windows
	allow and as far as the tracklist goes.
	When there are no previous tracks we show them empty for now.
	The user can scroll within the shown list.
	
	So we want to be informed when the current track has changed or when the tracklist has changed.
	
	We tell the model what we want with tracklist_range_set(start,end)
	Whenever the contents of the given tracklist range have changed, we are informed 
	by calling our view_tracklist_change. We could be given parameters telling us which range has changed
	to allow efficient uodating of our windows.
	
	We are further informed whenever the current track changes, because we show the current track differently.
	
	The variable first_shown_track is the number of the first track that we show.
	The variable last_shown_track is the number of the last track that we show.
*/	


/* ------------------------------------- Track List Menu ------------------------------------------ */
/* The row where the window list starts */
#define WL_START_ROW 0

/* The width of each window in pixels */
#define WL_WIDTH 128

/* The height of a small window */
#define WL_SMALL_HEIGHT 10

/* The height of a non-highlighted window in the window list */
#define WL_NORMAL_HEIGHT 13

/* The height of a highlighted window in the window list */
#define WL_HIGH_HEIGHT 18

/* Number of entries in the window list 
	We have 160 pixels, so 10 normal windows and 1 big window fit 
*/
#define WL_SIZE 12

/* The window list */
static struct Window win[WL_SIZE];

#define WIN_TXT_SIZE 64
/* Text buffer for the windows */
static char win_txt[WL_SIZE][WIN_TXT_SIZE];

/* This window is currently selected by the user */
static int sel_win;

/* The first track in our window list */
int first_track;

/* The last track in our window list */
// TODO hey we do not really know this
//int last_track;
#define last_track (first_track + WL_SIZE - 1)

/* These variables are set if we detect that the user wants something */
static enum USER_CMD user_cmd;
static int user_arg;


/* We need a function to get the song number corresponding to a window index (1..WL_SIZE-1) */
static int
track_no(int win_idx){
	int no;
	
	no = win_idx - 1 + first_track;
	if ((no < 0) || (no > last_track))
		return -1;
	return no;  
};

/* redraw all the windows 
	assumes the screen has just been cleared
*/
static void
tracklist_screen_redraw(){
	int i;
	
	for (i=0; i < WL_SIZE; i++)
		win_redraw(&win[i]);
};

static void
tracklist_screen_visible(int v){
	int i;
	for (i=0; i < WL_SIZE; i++)
		if (v) 
			win[i].flags |= WINFLG_VISIBLE;
		else
			win[i].flags &= ~WINFLG_VISIBLE;
};

static void
sel_win_up(){
	if (sel_win > 1){

		win[sel_win].font = MEDIUMFONT;
		win[sel_win].height = WL_NORMAL_HEIGHT;
		win[sel_win].start_row = win[sel_win-1].start_row + WL_HIGH_HEIGHT;
		win[sel_win].border = 0;
		win_unscroll(&win[sel_win]);
		sel_win--;						
		win[sel_win].font = BIGFONT;
		win[sel_win].height = WL_HIGH_HEIGHT;
		win[sel_win].border = 1;
		win_scroll(&win[sel_win]);				
		win_redraw(&win[sel_win]);
		win_redraw(&win[sel_win+1]);
	};
};

static void
sel_win_down(){
	if (sel_win < WL_SIZE - 1){
		win[sel_win].font = MEDIUMFONT;
		win[sel_win].height = WL_NORMAL_HEIGHT;
		win[sel_win].border = 0;
		win_unscroll(&win[sel_win]);
		sel_win++;		
		win[sel_win].font = BIGFONT;
		win[sel_win].height = WL_HIGH_HEIGHT;
		win[sel_win].border = 1;
		win[sel_win].start_row = win[sel_win-1].start_row + WL_NORMAL_HEIGHT;
		win_scroll(&win[sel_win]);
		win_redraw(&win[sel_win]);
		win_redraw(&win[sel_win-1]);
	};
};



/* We are called when the current song has changed. We can then adjust the range of our windows. */
void
view_pos_changed(int newpos){
	
// TODO at the moment  do not know what to do with this information
#if 0
	/* We try to get the current song, if there is any 
		because we want to adjust our track list range screen with the current song in the selected window.
	*/
	cur_song = newpos;
	if (cur_song < 0)	// either unknown or no current song
		cur_song=0;		// we default to the start of the playing list (if any)
	
	/* We like to start our list with the last 2 songs before the current one. */
	if (cur_song >= 2)
		first_track = cur_song - 2;
	else
		if (cur_song == 1)
			first_track = cur_song - 1;		// but we are content with only 1 song history or no history at all.
#endif		 
};


/* NOTE at the moment we live with the tracklist_changed() update, because a change of current song
	(nearly ?) always involves a change in the tracklist
*/

/* We are called when the current tracklist has potentially changed 
	TODO	Here we are updating the whole list. Could be inefficient.
		Better to update one song at a time.
*/
void
view_tracklist_changed(){
	int i;
	char *info;
/* TODO We are currently ignoring this ! */	
#if 0	
	/* We try to get the current song, if there is any */
	cur_song = mpd_get_song();
	if (cur_song >= 0)
		first_track = cur_song -2;
	if (first_track < 0)
		first_track = 0;
#endif
//	first_track = 15;	
		
	for (i=1; i < WL_SIZE; i++){
		info = track_info( track_no(i));
		if (NULL == info)
			win_new_text(&win[i], "");
		else win_new_text(&win[i], info);
	};

};



void
tracklist_screen_enter(){
	int cur_song;
	
	/* This is a sensible default */
	first_track = 0;
		
	/* We try to get the current song, if there is any 
		because we want to adjust our track list range screen with the current song in the selected window.
	*/
	cur_song = mpd_get_song();
	if (cur_song < 0)	// either unknown or no current song
		cur_song=0;		// we default to the start of the playing list (if any)
	
	/* We like to start our list with the last 2 songs before the current one. */
	if (cur_song >= 2)
		first_track = cur_song - 2;

	/* We always want to startup showing the current song in the selected window */
	while ( (sel_win + first_track - 1) < cur_song)
		sel_win_down();
		
	while ( (sel_win + first_track - 1) > cur_song)
		sel_win_up();

	// Tell model which range we want
	
	tracklist_range_set(first_track, last_track);
	
	lcd_fill(0x00);	
	
	/* A lot of things might have changed since we were called last.
		So better reread the information from mpd 
	*/
	view_tracklist_changed();
	
	tracklist_screen_visible(1);
	// TODO this is potentially time consuming. Do it piecewise (in a thread).
	tracklist_screen_redraw();		
};

void
tracklist_screen_exit(){
	tracklist_screen_visible(0);
};

void
tracklist_screen_init(){
	int i;
	int cur_start_row;
	
	user_cmd = NO_CMD;
	cur_start_row = WL_START_ROW;	
	
	/* The first window is a help window */
	win_init(&win[0], cur_start_row, 0, WL_SMALL_HEIGHT, 128,0, win_txt[0]);
	win[0].txt = win_txt[0];
	win[0].font = SMALLFONT;
	win[0].fg_color = WHITE;
	win[0].bg_color = BLACK;
	win[0].flags |= WINFLG_CENTER;
	cur_start_row += WL_SMALL_HEIGHT;
	
	sel_win = 3;	
		
	/* The remaining windows are track windows */
	for (i=1; i < WL_SIZE; i++){
		if (i==sel_win){
			/* The selected window is bigger than usual */ 	
			win_init(&win[i], cur_start_row, 0, WL_HIGH_HEIGHT, 128, 1, win_txt[i]);
			win[i].font = BIGFONT;
			cur_start_row += WL_HIGH_HEIGHT;	
		} else {
			win_init(&win[i], cur_start_row, 0, WL_NORMAL_HEIGHT, 128, 0, win_txt[i]);
			cur_start_row += WL_NORMAL_HEIGHT;
		};	
		/* Every second window has a different background color */
		if (0 == (i & 1)) win[i].bg_color = LIGHT_GREY;
	};
	
	win_new_text(&win[0], "Choose a track");
	
	win_scroll(&win[sel_win]);
	
};	


/* Returns user command if there is any, else returns NO_CMD */
enum USER_CMD 
tracklist_user_request(void){
	enum USER_CMD res = user_cmd;
	user_cmd = NO_CMD;
	return (res);
};

/* Sometimes we need an argument for the user cmd, this is returned by this function */
int 
tracklist_arg(void){
	return user_arg;
};


void
tracklist_keypress(int cur_key){
	switch (cur_key) {

		case KEY_OK:
			user_cmd = SEL_SONG;
			user_arg = track_no(sel_win);
			switch_screen(TRACKLIST_SCREEN, PLAYING_SCREEN);
			break;
		
		case KEY_Exit:
			switch_screen(TRACKLIST_SCREEN, PLAYLIST_SCREEN);
			break;
		
			
		case KEY_Up:
			if (sel_win > 1)
				sel_win_up();
			else {							// We want to scroll past the upper end of the window list
				if (track_no(1) > 0){		// We have information to show
					first_track--;
					tracklist_range_set(first_track, last_track);
					view_tracklist_changed();
				};
			};	
			break;
			
		case KEY_Down:
			if (sel_win < WL_SIZE - 1)
				sel_win_down();
			else{					// We want to scroll past the lower end of the window list
									// TODO here we should check if we reached the end of the playlist
				first_track++;
				tracklist_range_set(first_track, last_track);
				view_tracklist_changed();
			};
			break;
		
		case KEY_Info:
			break;		// We prevent a punch through 
	
			/* Punch Through the unknown keys */	
		default:
			playing_keypress(cur_key);
			break;
	};
};



