 /*
    window.c - window handling and drawing
	Copyright (C) 2009 by H. Raap

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "window.h"
#include "fonty.h"
#include "keyboard.h"
#include "lcd.h"
#include "timerirq.h"
#include "mpd.h"
#include "screen.h"

/* 
These are the window manager functions.
The window manager knows all about the current layout of the screen and handles it. 

The windows are organized in a list (an array).
So we may do some operations for all windows by iterating through this list.
Might as well be a linked list or something.

*/

#define WIN_DEFAULT_TXT_SIZE 80

void
win_draw_border(struct Window *win){
	if ( (win->border) && (win->flags & WINFLG_VISIBLE) )
		draw_rect(win->start_col, win->start_row, win->width, win->height, 1, BLACK);	
};
	

/* 
	The slot win.border is either 0 or 1.
	If it is 0, we have no visible border, else we have a 1 pixel border at all 4 sides.
	
	The text is drawn 2 pixels away from the left and right edges of the window.
*/ 


/* Initializes a window.  
	Each window is NOT visible by default and is assumed to be a TEXT type window
*/
void 
win_init(struct Window *win, uint8 row, uint8 col, uint8 height, uint8 width, uint8 border, char *txt){
	win->start_row = row;
	win->start_col = col;
	win->height = height;
	win->width = width;
	win->fg_color = BLACK;
	win->bg_color = WHITE;
	win->border = border;
	win->flags = WINFLG_TEXT;
	win->fits = 1;
	win->scroll = 0;
	win->font = MEDIUMFONT;
	win->buffer_lim = WIN_DEFAULT_TXT_SIZE;
	win->cur_char = 0;
	win->txt = txt;
	txt[0] = '\0';
	win->txt_offs_row = 1;
	win->txt_offs_col = 2;
	win->cur_col = win_txt_col(win);

};

/* Sets visible content of window to background color.
	If clr_border is 1, the border (if any) will also be cleared.
*/
void 
win_clear(struct Window *win, int clr_border){
	int border = win->border;
	if (! (win->flags & WINFLG_VISIBLE)) return;
	if (clr_border) border = 0;
	
	draw_block(win->start_row+border, win->start_col+border, win->width-2*border, win->height-2*border, win->bg_color);
};


static uint8 
draw_cur_char(struct Window *win){
	uint8 res;
	char ch;
	
	/* If we are past the text length, draw space */
	if (win->cur_char >= win->text_len) ch = ' ';
	else ch = win->txt[win->cur_char];
	
	res = draw_char(win_txt_row(win), win->cur_col, ch, win->fg_color, win->bg_color,  txt_col_lim(win) - win->cur_col ); 
	
	// if it does not fit completely, return 0.
	if (res == 0) return 0;
	
	// if it fits completely in remaining window space,
	// increase last character and last character column
	win->cur_col += res;
	
	/* After we have drawn 4 spaces beyond the current length, we continue with the beginning of the text */
	if (++win->cur_char >= win->text_len + 4)
		win->cur_char = 0;
	
	return 1;
};


/* 	Clear window (except border) and draw the contents completely (except border) new. Use text in win->buffer 
	Sets win->fits to 1 if the text fits in window, else sets it to 0.
	Text is centered in window (horizontally as well as vertically)
*/
static void
win_center_text(struct Window *win){
	int txt_height, txt_offs, cur_row;
		

	/* center vertically */
	txt_height = draw_text_space(win->txt, win_txt_row(win), win->start_col + win->border, 
								 win->width - 2*win->border, win->height - 2*win->border, win->fg_color, win->bg_color, 1);

	txt_offs = ( win_txt_height(win) - txt_height) / 2;
	
	if (txt_offs >= 0)
		cur_row  = win_txt_row(win) + txt_offs;
	else 
		cur_row  = win_txt_row(win);
	
	 draw_text_space(win->txt, cur_row, win->start_col + win->border, 
					 win->width - 2*win->border, win->height - 2*win->border, win->fg_color, win->bg_color, 0);

};

/* 	Clear window (except border) and draw the contents completely (except border) new. Use text in win->buffer 
	Sets win->fits to 1 if the text fits in window, else sets it to 0.
*/
void 
win_draw_text(struct Window *win){
	if (! (win->flags & WINFLG_VISIBLE)) return;
	if (win->txt == NULL) return; 	
	win_clear(win, 0);
	set_font(win->font);
	
	if (win->flags & WINFLG_CENTER)
		return win_center_text(win);
	
	win->cur_char = 0;
	win->cur_col = win_txt_col(win);
	win->fits = 1;
	while (win->cur_char < win->text_len){
		if (draw_cur_char(win) == 0) {
			win->fits = 0;
			break;
		};
	};
};
		

/* We are given a new text 
	If it is different from current text in window, 
	copy text to window and redraw window
*/
void 
win_new_text(struct Window *win, char *s){
	/* We copy the new text to our internal buffer with clipping if necessary
		We also check if the text currently in buffer is identical
	*/
	if ( strn_cpy_cmp(win->txt, s, win->buffer_lim - 1, &(win->text_len)) != 0)
		return;			// the two strings are identical, do nothing

	win_draw_text(win);
};

void
win_redraw(struct Window *win){
//	win_clear(win, 1);				// draw text clears the window 
	if (! (win->flags & WINFLG_VISIBLE)) return;
	win_draw_border(win);
	if (win->flags & WINFLG_TEXT)
		win_draw_text(win);
	else if (win->flags & WINFLG_BAR){
		draw_progressbar(win->start_row+1, win->start_col, win->height-2, win->fg_color, win->bg_color, 100, 0);
		draw_progressbar(win->start_row+1, win->start_col, win->height-2, win->fg_color, win->bg_color, 0, win->cur_char);
	} else if (win->flags & WINFLG_RAMP){			
			trianglebar_border(win->start_row+1, win->start_col+1, win->width-2, win->height-2, win->fg_color, win->bg_color);
			draw_trianglebar(win->start_row+1, win->start_col+1, win->width-2, win->height-2, win->fg_color, win->bg_color, 0, win->cur_char);
	};
};

/*
	Draw the current character at the current character column
	If it did fit completey in window, we advance current character by 1
	Returns 0, if character does not fit completely in window,
	else returns 1.
	Assumes that the font is already set.
*/


/* ### Scroll Task ###  
	Some windows might need scrolling to show their full content.
	This task regularily (after a timer has expired) checks each window in turn if it needs scrolling.
	If so, the window is scrolled and the task yields.
	If not, we wait for next timer expiration.
	So we make sure that at most 1 window is scrolled before we are yielding to minimize the time spent in this task.

	TODO If the font of a window has changed, check again if the text fits!

	We do not check all available windows, that would be a logistic nightmare.
	We keep a small list (say 10) of windows to check.
	If you want a window to scroll, add it with win_scroll() and remove it with win_unscroll().
*/
#define SCROLL_LIST_MAX 10
static struct Window *scroll_list[SCROLL_LIST_MAX + 1];

#define SCROLL_OFFSET 8
static void 
win_do_scroll(struct Window *win){ 	
	int res;
	
	set_font(win->font);
	
	scroll(win_txt_row(win), win_txt_col(win), win_txt_width(win), SCROLL_OFFSET);
	
	if (win->cur_col - SCROLL_OFFSET >= 0)
		 win->cur_col -= SCROLL_OFFSET;
	else
		win->cur_col = 0;
	
	do {
		res = draw_cur_char(win);
	} while (res);
};	


void 
win_scroll(struct Window *win){
	int i;
	for (i=0; i < SCROLL_LIST_MAX; i++){
		if (scroll_list[i] == NULL){
			scroll_list[i] = win;
			return;
		};
	};
};
	
void 
win_unscroll(struct Window *win){
	int i;
	for (i=0; i < SCROLL_LIST_MAX; i++){
		if (scroll_list[i] == win){
			scroll_list[i] = NULL;
			return;
		};
	};
};

#define SCROLL_PERIOD (8 * (TICKS_PER_TENTH_SEC))
PT_THREAD (win_scroll_all(struct pt *pt)){
	static int i;
	static struct timer scroll_tmr;
	
	PT_BEGIN(pt);
		
	timer_add(&scroll_tmr, SCROLL_PERIOD, SCROLL_PERIOD);
	while(1){
		for (i=0; i < SCROLL_LIST_MAX; i++){
			PT_WAIT_UNTIL(pt, timer_expired(&scroll_tmr));
			if (scroll_list[i] == NULL) continue;
			if (! (scroll_list[i]->flags & WINFLG_VISIBLE)) continue; 
			if ( ! scroll_list[i]->fits ){
				win_do_scroll(scroll_list[i]);
				PT_YIELD(pt);
			};
		};
		/* We have checked (and potentially scrolled each window). Timer can expire again. */
		scroll_tmr.expired=0;
	};
	PT_END(pt);
};

void
win_scroll_init(){
	int i;
	for (i=0; i < SCROLL_LIST_MAX; i++)
		scroll_list[i] = NULL;
	
	task_add(&win_scroll_all);
};	


#if 0
// TODO is this called anywhere ? 
void 
new_rtc_time(unsigned char *t) {
	char timetext[26];
	
	time2string(t, timetext);
//	win_new_text(&win[5], timetext);
};
#endif

#if 0
/* TODO this function should be called win_new_text */
/* We are informed that the window text of window pos may have changed */
void
win_changed(int pos, char *s){
	if ((pos < 0) || (pos > 4)) return; 
	win_new_text(&win[pos], s);
};
#endif

