/* model.c 
	Copyright (C) 2009 by H. Raap

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Information about status is stored in this structure. 
	We need to indicate that we do not know some things in here.
	For volume, song and song_id we use -1 to indicate UNKNOWN.
	The state has its own UNKNOWN value.
	cur_artist ard cur_title are NULL if UNKNOWN.
	
	We make sure that the information here is consistent, i.e. if for example songid is set,
	the variables pos, time_total etc. reflect the information about that particular song and not
	about some earlier song.
	
*/
#include "global.h"
#include "fonty.h"
#include "mpd.h"
#include "model.h"

// Length of artist and title strings, some songs really have long titles
#define STRLEN 150


// TODO just a hack !!! remove it
static const char EMPTY_STRING[1] = "";


struct STATUS {
	int playlistlength;			// -1 if unknown
	int volume;					// -1 if unknown
	int old_volume;				// -1 if unknown, volume before a mute command, used to unmute
	int pos; 					// current song position in the playlist. Starts with 0!
	enum PLAYSTATE state;
	int songid;					// mpd's internal id of the current song
	char artist_buf[STRLEN];	// If cur_artist points here, this is the artist tag, else irrelevant 
	char title_buf[STRLEN];		// If cur_title points here, this is the title tag, else irrelevant 
	char *cur_artist;			// NULL if not known, else points to info in artist_buf;
	char *cur_title;			// NULL if not known, else points to info in title_buf;
	int time_elapsed;			// in seconds
	int time_total;				// in seconds
	int cur_playlist;			// index to currently loaded playlist (or -1), not used in mpd_status, only in user_status
};

/* The currently known status of mpd. */
static struct STATUS mpd_status;

/* The status as the user wants it to be */
static struct STATUS user_status;

static uint32 model_changed_flags;			// Flags which are set when (part of) the model has changed

/* ----- Playlist variables, keeping info about playlists known to us. ----- */
/* Maximum length of a playlist name including trailing 0 */
#define PL_NAME_LENGTH 32

/* Maxumum number of playlist names that we want to know */
#define PL_LIM 20

/* Maximum index into pl_name[] */
#define PL_MAX (PL_LIM - 1)

/* The names of all the known playlists */
static char pl_name_entry[PL_LIM][PL_NAME_LENGTH];

/* Pointers to the playlist names or NULL if unknown */
static char *pl_name[PL_LIM];

/* Number of playlist names currently known to us. */
static int pl_num;

/* ----- Tracklist variables, keeping info about songs in the current playlist. ----- */

/* Number of tracklist entries that we do store */
#define TL_LIM 20

/* Maximum index into our tracklist array */
#define TL_MAX (TL_LIM - 1)

/* Concatenation of Title and Artist tag in the current playlist 
	This is a ring buffer.
	The buffer starts at tl_first_entry and contains TL_LIM entries.
	Not all entries may be filled with information.
	The first entry (at tl_first_entry) corresponds to track number tl_begin in the current playlist.
*/
static char tracklist_entry[TL_LIM][TL_ENTRY_LENGTH];

/* Our tracklist portentially contains infos about tracks with track no. from tl_begin to tl_end */
static int tl_begin;

/* Last track number in our tracklist */
#define tl_end (tl_begin + TL_MAX)

/* At which position does our list start, i.e. which index corresponds to tl_begin? */
static int tl_first_idx;

/* Pointers to tracklist_entries 
	We do this indirection to allow fast scrolling of the list and to set a NULL value for unknown values.
*/
static char *tracklist[TL_LIM];

/* 
	We try to know how good our communication with mpd is going. 
	So we keep a record of errors.  See below for a better description of error handling. 
*/
static int comm_error;


/* ------------------------------------- Retrieving information from the model ---------------------------------- */
/* The error count. Between 0 and 255. Higher means more errors occured. */
int
get_comm_error(){
	return comm_error;
};

int
model_get_changed(){
	return model_changed_flags;
};


/* Used by tracklist screen to know which song is current 
	NOTE here we might return -1 or -2 if there is no current song or it is unknown.
		caller has to handle those values
*/
int
mpd_get_song(){
	return mpd_status.pos;
};

/* 
	Might return -1 if we do not know the current playlistlength.
	Returns 0 if there is no song in the playlist (maybe just cleared)
*/
int
mpd_get_playlistlength(){
	return mpd_status.playlistlength;
};

int
mpd_get_state(){
	return mpd_status.state;
};

/* Returns a string with the current title or NULL if unknown. Tracks with no title return "". */
char *
mpd_get_curtitle(){
	return mpd_status.cur_title;
};

/* Returns a string with the current artist or NULL if unknown. Tracks with no artist return "". */
char *
mpd_get_curartist(){
	return mpd_status.cur_artist;
};



/* NOTE here we ask for information from user_state. Why that? It's not logical. */

int
user_get_state(){
	return user_status.state;
};

int	
mpd_get_volume(){
	return mpd_status.volume;
};

int	
user_get_pos(){
	return user_status.pos;
};

int
mpd_get_time_elapsed(){
	return mpd_status.time_elapsed;
};

int
mpd_get_time_total(){
	return mpd_status.time_total;
};

/* Given an index starting from 0, we return the corresponding playlist name entry.
	If we have no info we return "".
 	If the playlist does not exist, we return NULL.
*/
char *
pl_get_name(int idx){
	if ( (idx < 0) || (idx >= pl_num) ) 
		return NULL;
	if (pl_name[idx] == NULL) 
		return "";
	return pl_name[idx];
};

	
/* Given a track no, we return the index into tracklist[] corresponding to this number.
	If the number is outside of our list, we return -1.
*/
static int
tl_index(int track_no){
	int idx;

	if ( (track_no < tl_begin) || (track_no > tl_end) )
		return -1;
		
	idx = tl_first_idx + track_no - tl_begin;
	if (idx >= TL_LIM)
		idx -= TL_LIM;	
	return idx;
};

/* We are given a track number. We return the corresponding tracklist entry 
	or NULL if the track is outside our list.
	NULL might also mean we do not have information about that track.
*/
static char *
tl_get_entry(int track_no){
	int i = tl_index(track_no);
	if (i < 0) return NULL;
	return tracklist[i];
};

/* The tracklist screen will ask us for the track entry of a specific song no. 
	We return it if available, else we return either "" if that track does not exist in the current playlist
	or NULL if we know nothing about this track.
	The track numbering starts with 0 as mpd does.
	NOTE This is the only information the outside world gets about our tracklist.
		A scrolling client may want to know if he has reached the end of the playlist.
		This routine does not do that.
		Either return NULL only if that occurs and maybe "?" or so for unknown values
		or have a seperate function like bool track_exists(int no).
*/
char *
track_info(int no){
	
	/* If we know nothing about the current playlist, we return nothing. */
	// NOTE not logical. Even if we do not know the playlistlength, we might have info in our tracklist
	if (mpd_status.playlistlength == -1)
		return NULL;

	/* If the song is outside the current playlist, we return "" (track does not exist)  */
	if ( (no < 0) || (no >= mpd_status.playlistlength) )
		return "";
	
	/* Now we give the real info */
	return tl_get_entry(no);
};


/* Returns track-no if we need info about a certain song for our tracklist, else -1. */ 
static int 
need_tracklist(){
	int i, search_end;

	if (mpd_status.playlistlength >= 0)
		search_end = min(tl_end, mpd_status.playlistlength - 1);
	else
		search_end = tl_end;

	/* Search through our tracklist. Returns first track which has no info */
	for (i=tl_begin; i <= search_end; i++)
		if (tl_get_entry(i) == NULL)
			return i;
	return -1;
};


/* Returns TRUE iff we need to know the artist or the title of the current song. */
static inline int 
need_artist_or_title(){
	return ((mpd_status.cur_artist == NULL) || (mpd_status.cur_title == NULL));
};

/* Returns TRUE iff there are one or more variables in mpd_status that are unknown and which we can get
	be issuing a status command.
*/
static int
need_status(){
	/* If the following variables are unknown, we must definitely issue a STATUS command */
	if ( (mpd_status.playlistlength == -1) || (mpd_status.volume == -1) || (mpd_status.state == UNKNOWN) ) 
		return 1;
	
	/* We use status when we want to know the current song number, but that only makes sense when we assume there is a current song. */
	if ( (mpd_status.playlistlength > 0) && (mpd_status.pos == SONG_UNKNOWN) )
		return 1;
	
	/* The remaining values are not given to us, when there is no current song */
	if (mpd_status.pos == NO_SONG) return 0;
		
	/* MPD gives us no time information, if the state is stopped. Stupid, but true. */
	if (mpd_status.state == STOP) return 0;
	
	return ( (mpd_status.time_elapsed == -1)
			|| (mpd_status.time_total == -1)
		  );		
};


/* Here we determine if user_status and mpd_status agree and if we need any information from mpd 
	If the command has an argument, we return it in *arg
	We return either a cmd or NO_CMD.
	The actions here are sorted according to importance, most important first, because we just return
	1 action, the most important one.
*/
enum USER_CMD 
model_needs_action(int *arg){	
	int tmp;
	/* mpd playing state and the user state should match */
	if ( (user_status.state != UNKNOWN) && (mpd_status.state != user_status.state) ){
		if (user_status.state == PLAY){
			/* The user definitively wants to play something */
			if (mpd_status.state == PAUSE)
				return (PAUSE_OFF);
			/* Only issue a play command if the playlist is not empty 
				We might just have issued a CLEAR command, which stopped MPD, but user wants playing.
				Issue PLAY command after we have loaded the new playlist and determined it is not empty.
			*/
			if (mpd_status.playlistlength > 0) 
				return (PLAY_CMD);
		};
		
		if (user_status.state == PAUSE)
			return (PAUSE_ON);
		
		if (user_status.state == STOP)
			return (STOP_CMD);
	};
		
	/* Does the user want a specific volume that is not currently set ? */
	if ( (user_status.volume != -1) && (user_status.volume != mpd_status.volume)){
		*arg = user_status.volume;
		return VOLUME_NEW;
	};
	
	/* Maybe the user selected a specific song to play now */
	if ((user_status.pos >= 0) && (user_status.pos != mpd_status.pos)){
		*arg = user_status.pos;
		return SEL_SONG;
	};
	
	/* We ask for status before asking for current song, because after status we know if there is any current song */
	if (need_status())
		return (STATUS_CMD);

	if (need_artist_or_title())
		return (CUR_SONG_CMD);
	
	if (user_status.time_elapsed > mpd_status.time_elapsed ){
		*arg = user_status.time_elapsed;
		return (FORWARD_CMD);
	};
	
	if ( (user_status.time_elapsed >= 0) & (user_status.time_elapsed < mpd_status.time_elapsed )){
		*arg = user_status.time_elapsed;
		return (REWIND_CMD);	
	};
	
	tmp = need_tracklist();		/* Maybe costly operation. Cache its value. */
	if (tmp >= 0){
		*arg = tmp;
		return (PLINFO_CMD);
	};	
	
	if (user_status.playlistlength == 0)
		return (CLEAR_CMD);
	
	if (user_status.cur_playlist >= 0){
		/* It would be more logical to return a char* here, but easier to return the index to it. */
		*arg = user_status.cur_playlist;
		return (LOAD_CMD);
	};
	
	/* TODO the available playlists may change by an outside action (another client created/deleted one etc.)
			So we should regularily (every 5 minutes or so) reread this information
	*/ 
	if ( pl_num < 0 )					/* We have no idea about available playlists */
		return (LSINFO_CMD);

	
	// TODO here we should check if we have had a long time without a new status.
	//		Then simply do a "STATUS" command to synchronize MPD and this model.
	//		We have to record when the last status command was successfully executed.
	
	return NO_CMD; 
};


/* ------------------------------------- Changing information in the model -------------------------------------- */

/* Setting flags indicating something has changed */
static inline void
model_changed(uint32 change){
	model_changed_flags |= change;	
};

void
model_reset_changed(){
	model_changed_flags = 0;
};

// TODO put this in a module global.c together with other helper routines
static int abs(int a){
	if (a >= 0) return a;
	else return -a;
};

/* We have (new) timing information from mpd */
void
mpd_set_time(int elapsed, int total){
	if ( (mpd_status.time_elapsed != elapsed) || (mpd_status.time_total != total) )
		model_changed(TIME_CHANGED);
	mpd_status.time_elapsed = elapsed;
	mpd_status.time_total = total;
	/* Here we should check if the user really cares for a specific time.
		If it is closely within the real time, we set users wish to don't care,
		else we leave it as it is (and later try again to skip forward or backward)
	*/
	if (abs(user_status.time_elapsed - mpd_status.time_elapsed) < 3)
		user_status.time_elapsed = -1;		// wish fulfilled
};


/* Time flies by when you're the driver of a train :-)
	We increment the elapsed time by 1 second if possible.
	This is done because we expect mpd to behave nicely.
	TODO Maybe add a task somewhere that synchronizes the suspected status (incl. time) with the status as returned by mpd
	in regular intervals (30 secs or so).
*/
void
mpd_inc_time(){
	/* Only increase time if we are playing something */
	if (mpd_status.state == PLAY){
		if (mpd_status.time_elapsed >= 0){		// We have a time that we can increment 

			/* If time is up, we set the mpd values to -1 to tell controller that we need info */
			if (mpd_status.time_elapsed >= mpd_status.time_total){
				/* We assume the current song has ended playing.
					We then don't know very much, because playlist might have ended. */
				mpd_set_pos(SONG_UNKNOWN);
			} else {
				/* Time is not up. We can increment elapsed time */
				mpd_status.time_elapsed++;
				model_changed(TIME_CHANGED);
			};
		};
	};
};


void
mpd_set_state(enum PLAYSTATE newstate){
	if (newstate == mpd_status.state) return;
	mpd_status.state = newstate;		
	
	/* This is important to do, because sometimes mpd changes state itself, for example when the playlist finished.
		Then we do not want to repeatedly issue a useless "PLAY" command because the user some time back wanted to play */
	if (mpd_status.state == user_status.state)		// wish fulfilled
		user_status.state = -1;
	
	/* We only set the model_changed flags when we know the status, because else we have nothing to show the user.
		We will get a definitive playstate later with a status command (we hope so).
	*/
	if (newstate != UNKNOWN)
		model_changed(STATE_CHANGED);
};

/* We are given either NULL or a string with the artist tag, NULL meaning artist unknown */
// TODO how can artist unknown happen ? Should we better store an empty string then ?
void
mpd_set_artist(char *s){
	int length;
	
	if ( (s == NULL) ){
		if (mpd_status.cur_artist != NULL){
			mpd_status.cur_artist = NULL;		
			model_changed(ARTIST_CHANGED);
		};
	} else {
		if (strn_cpy_cmp(mpd_status.artist_buf, s, STRLEN, &length) == 0)	
			model_changed(ARTIST_CHANGED);
		mpd_status.cur_artist = mpd_status.artist_buf;
	}; 
};

/* We are given either NULL or a string with the title tag, NULL meaning title unknown */
// TODO how can title unknown happen ? Should we better store an empty string then ?
void
mpd_set_title(char *s){
	int length;
	
	if (s == NULL){
		if (mpd_status.cur_title != NULL){
			mpd_status.cur_title = NULL;
			model_changed(TITLE_CHANGED);
		};
	} else {
		if (strn_cpy_cmp(mpd_status.title_buf, s, STRLEN, &length) == 0)	
			model_changed(TITLE_CHANGED);
		mpd_status.cur_title = mpd_status.title_buf;
	}; 
};

/* We have got (maybe new) information about the current song pos */ 
void
mpd_set_pos(int newpos){
	if (mpd_status.pos == newpos) return;		// nothing new
	mpd_status.pos = newpos;

	if (mpd_status.pos == user_status.pos)		// wish fulfilled
		user_status.pos = -1;
	
	if (newpos == NO_SONG){ 
		mpd_set_artist("");
		mpd_set_title("");
		mpd_status.songid = -1;				// NOTE songid is currently not used
		mpd_set_time(-1, -1);	
		mpd_set_state(STOP);
	} else { 
		mpd_set_artist(NULL);				/* We do not yet know which artist and title we have */
		mpd_set_title(NULL);
		mpd_status.songid = -1;				// NOTE songid is currently not used
		mpd_set_time(0, -1);				/* The 0 is just a guess. But we have to get total time anyway. */
	};
	model_changed(POS_CHANGED);				// this is (maybe) important to know for track list handling */
};


/* The length of our playlist has changed. We use the given number if it is >= 0.
	Else it means we do not know the current playlistlength (maybe no playlist at all).
	We adjust the information in our tracklist accordingly and maybe current_song as well.

	TODO we really need a function that is different from this which is called clr_playlist
		because we try to do here more than we know.
*/
void
set_playlistlength(int n){
	if (n < 0) {
		mpd_status.playlistlength = n;
		return;
	};
	if (user_status.playlistlength == n)
		user_status.playlistlength = -1;							// wish fulfilled
	
	if (mpd_status.playlistlength == n)
		return;														// no change
	
	mpd_status.playlistlength = n;	
	
	/* If the playlist is empty, there can be no current song.*/
	if (n == 0){
		mpd_set_pos(NO_SONG);
	};
	// TODO Here we should issue TL_CHANGED flags, if appropriate. */
};

/* We have some information about the current mpd volume.
	Update mpd_status and if necessary user_status
*/
void
mpd_set_volume(int vol){
	if (vol != mpd_status.volume){
		mpd_status.volume = vol;
		model_changed(VOLUME_CHANGED);
	}
};


/* Given title, artist and track number, store this info in our tracklist (if it fits) */
void
tl_store_track(char *title, char *artist, int track_no){
	int idx;
	char *entry;
	
	idx = tl_index(track_no);
	if (idx < 0)
		return;

	entry = tracklist_entry[idx];
	
	strn_cpy(entry, title, STRLEN); 
	str_cat_max(entry, " - ", STRLEN);
	str_cat_max(entry, artist, STRLEN);
	tracklist[idx] = entry;
	model_changed(TRACKLIST_CHANGED);
};

/* All the songs starting with number no have been changed and are currently unknown. */
void 
tracks_unknown(int no){
	int i;
	
	if (no > tl_end) 
		return;				/* not in our model anyway */
	if (no < tl_begin)
		no = tl_begin;

	for (i=no; i <= tl_end; i++)
		tracklist[tl_index(i)] = NULL;
};
		
/* We are shifting the start of our tracklist up by 1 song 
	We are not changing most of the information in tracklist[], only the start (and end) of the array.
	We are setting the information of the previous first track which is now the new last track no.
	to NULL, because we do not know it.
*/
static void
tl_shift_up_1(){
	tracklist[tl_first_idx] = NULL;
	tl_begin++;
	tl_first_idx ++;
	if (tl_first_idx > TL_MAX)
		tl_first_idx = 0;
};

/* We are shifting the start of our tracklist down by 1 song 
	We are not changing most of the information in tracklist[], only the start (and end) of the array.
	We are setting the information of the new first track no. to NULL, because we do not know it.
*/
static void
tl_shift_down_1(){
	tl_begin--;
	tl_first_idx--;
	if (tl_first_idx < 0)
		tl_first_idx = TL_MAX;
	tracklist[tl_first_idx] = NULL;
};

		

/* We are shifting the start of our tracklist up by diff songs.
	If we leave the current playlist range too much, all our information is worthless
	and we forget it.
	Else we keep that part of info that we already gathered.
*/
static void
tl_shift_up(int diff){
	if (diff > TL_MAX){
		tracks_unknown(tl_begin);		// all infos are invalid.
		tl_begin += diff;
	} else {
		for (;diff > 0; diff --)
			tl_shift_up_1();
	};
};

/* We are shifting the start of our tracklist down by diff songs.
	If we leave the current playlist range too much, all our information is worthless
	and we forget it.
	Else we keep that part of info that we already gathered.
*/
static void
tl_shift_down(int diff){
	if (diff > TL_MAX){
		tracks_unknown(tl_begin);		// all infos are invalid.
		tl_begin -= diff;
	} else {
		for (;diff > 0; diff --)
			tl_shift_down_1();
	};
};


/* The user has requested a certain range for our tracklist.
	Given are the first and last track that the user wants info to be available for.

	We adjust the range of our internal tracklist accordingly, trying to keep as much information
	valid as we can, so we do not have to reread everything.
*/
void
tracklist_range_set(int start, int end){
	if (start < 0 ) start = 0;		// sanity check, should not occur
		
	if (end > tl_end)			// user wants more information than we currently have
		tl_shift_up(end - tl_end);
	else
		if (start < tl_begin)		// user wants information from earlier tracks than we have
			tl_shift_down(tl_begin - start);
};


void
user_song_unknown(){
	user_status.pos = SONG_UNKNOWN;
	user_status.state = UNKNOWN;
	user_status.cur_artist=NULL;
	user_status.cur_title=NULL;
	user_status.songid = -1;
	user_status.time_elapsed = -1;
	user_status.time_total = -1;
};

/* This is called before controller begins getting playlist names from MPD. */
void
start_reading_playlists(){
	if (pl_num != 0){
		pl_num = 0;
		model_changed(PL_NAMES_CHANGED);
	}
};

// NOTE this is a bit tricky. We can be calledt multiple times and we really do not know
//		where to store the current playlist name, because playlists are not numbered in the current MPD.
//		We rely on the fact that before calling this routine, the variable pl_num has been reset to 0 !
void
mpd_set_playlistname(char *s){
	if (pl_num < PL_LIM){ 
		/* No use to store playlists with too long names, because we can then later not load the shortened names. */
		if (str_len(s) > PL_NAME_LENGTH-1) return;
		
		strn_cpy(pl_name_entry[pl_num], s, PL_NAME_LENGTH-1);
		pl_name[pl_num] = pl_name_entry[pl_num];
		
		model_changed(PL_NAMES_CHANGED);
		pl_num++;
	};
};	

/* Called after a successful LOAD command. We assume the user wish has been fulfilled 
	We do not care what playlist was loaded.
	But we assume the user wants to play the first song of the playlist.
*/
void
mpd_set_newplaylist(){
	user_status.cur_playlist = -1;
	
	/* After loading a playlist, MPD does not change its state, but the playlist has been expanded */
	/* If we knew the previous playlistlength, we know which tracks we must update */
		
	if (mpd_status.playlistlength >= 0)
			tracks_unknown(mpd_status.playlistlength);
		else
			tracks_unknown(0);		// update all tracks
	set_playlistlength(-1);
	user_status.pos = 0;		// or mpd_want_song(0)
	user_status.state = PLAY;
};	

/* ------------------------------------------------ Wishing functions ---------------------------------------------- */

void
user_set_state(enum PLAYSTATE s){
	user_status.state = s;
};

void
mpd_want_song(int p){
	/* We make sure that the wish makes sense. */
	if (p < 0) return;
	if ((mpd_status.playlistlength >= 0) && (p > mpd_status.playlistlength - 1))
				 return;
	
	user_status.pos = p;
};

/* User wants next or previous song, so offs should be 1 or -1. Does no harm if it is not. */
void
mpd_want_song_add(int offs){
	if (mpd_status.pos >= 0)
		mpd_want_song(mpd_status.pos + offs);
}

/* Changes the internal user volume by adding a value to mpd_status.volume 
	If mpd_status.volume is unknown at this point, wished volume will not be set.
*/
void
mpd_want_volume_add(int chg){
	int tmp;
	if (mpd_status.volume == -1) return;
	tmp = mpd_status.volume + chg;
	if (tmp < 0)
		tmp = 0;
	if (tmp > 100)
		tmp = 100;
	user_status.volume = tmp;
};


/* The user wants to skip forward or backward in time a few seconds. */
void
mpd_want_time_add(int offs){
	int tmp;
	if (mpd_status.time_elapsed == -1) return;
	tmp = mpd_status.time_elapsed + offs;
	if (tmp < 0) tmp = 0;
	if (mpd_status.time_total != -1)
		if (tmp > mpd_status.time_total) 
			tmp = mpd_status.time_total;
	user_status.time_elapsed = tmp;
};


/* User wants to toggle the mute status, i.e. if volume was 0, reset it to old value, else set it to 0. 
	volume == 0 is treated as MUTE 
*/
void
user_mute(){
	if ( mpd_status.volume != 0 ){
		mpd_status.old_volume = mpd_status.volume;		// remember the current volume
		user_status.volume = 0;
	} else
		if (mpd_status.old_volume > 0)
			user_status.volume = mpd_status.old_volume;
		else
			user_status.volume = 15;						// we use a default here in case we have no idea of old volume
};



/* The user wants a new playlist. 
	We store a reference to the resulting string, so its contents must be kept valid ! 
	We assume the user really wants two things.
	First he wants the current playlist to be emptied.
	Then he wants to load the new playlist.
	This is a bit different than MPD interprets the LOAD command.
	So we set two wishes here at once. 
	The need_action() routine above must make sure to handle the CLEAR request first.
*/
void 
mpd_want_playlist(int idx){
	if ((idx < 0) || (idx >= pl_num))
		return;
	
	user_status.playlistlength = 0;
	user_status.cur_playlist = idx;
	/* The user explicitly requested this, so we do a complete reload of the new playlist
		without checking if it is already loaded.
		We have no easy way to know which playlist is loaded, because mpd does no bookkeeping about playlist names.
		We can only detect that the  "LOAD" command worked by waiting for its "OK".
	*/
};



/* ----------------------------------------------- Initialization -------------------------------------------------- */
/* sets all values in st to UNKNOWN */
static void
init_status(struct STATUS *st){
	st->playlistlength = -1;
	st->old_volume = -1;
	st->volume = -1;
	st->state = UNKNOWN;	
	st->cur_playlist = -1;
};


void 
model_init(){
	int i;
	
	comm_error = 0;
	init_status(&mpd_status);
	mpd_set_pos(SONG_UNKNOWN);
	
	init_status(&user_status);
	user_song_unknown();
	
	pl_num = -1;		/* No info about playlists. */
	
	/* We assume that the user wants to play immideately */
	user_status.state = PLAY;
	
	/* We have just reset all infos. So the view has to be updated. */
	model_changed_flags = VOLUME_CHANGED | TITLE_CHANGED | ARTIST_CHANGED | TIME_CHANGED;
	
	tl_first_idx = 0;
	tl_begin = 0;
	
	for (i=0; i<TL_LIM; i++)
		tracklist[tl_index(i)] = NULL;
};

/* Some error has occured (maybe timeout) 
	We increase our error count.
*/
void 
error_inc(int amount){
	if (comm_error + amount > 255)
		comm_error = 255;
	else 
		comm_error += amount;
};


/* We had successful communication with mpd 
	We decrease our error count.
*/
void
error_dec(int amount){
	if (comm_error - amount < 0)
		comm_error = 0;
	else
		comm_error -= amount;
};


/* The user has hit the pause toggle key:
	We try to find out what state user wants to be in
*/
void
user_toggle_pause(){

/* We know the current state of mpd, so we assume the user wants to toggle that. */
		if (mpd_status.state != UNKNOWN){
			if ( (mpd_status.state == PAUSE)|| (mpd_status.state == STOP) )
				user_status.state = PLAY;
			if (mpd_status.state == PLAY)
				user_status.state = PAUSE;

			/* The pause key shall have no function if in any other mode. */
		} else {
			/* current mpd state is unknown, so we assume the user state is correct. 
				TODO NO, this is not correct. We simply have to know mpd_status_state first.
			*/
#if 0			
			if (user_status.state == PAUSE){
				user_status.state = PLAY;
				mpd_status.state = PAUSE;
			}
			else if (user_status.state == PLAY) {
				user_status.state = PAUSE;
				mpd_status.state = PLAY;
			};
#endif		
		};
};

