/*
    mpd.c - MPD control functions
	Copyright (C) 2009 by H. Raap

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/*
  We use a model-view-controller architecture here
  
	Our model consists of the status of mpd (inclusive (part of) the current playlist stored in the array play_list[])
	and a the status as the user wants it to be.
	If the user status and the mpd status differ, some action is needed (usually commands to mpd).
	
	The view are windows on screen, with 1 window which has the focus
	Changes in the model are reported to the view.
	
	The controller takes input from user and from mpd and changes the model
	accordingly.

*/

#include "global.h"
#include "kernel.h"

#include "timerirq.h"
#include "pt-sem.h"
#include "rf.h"
#include "fonty.h"
#include "window.h"
#include "screen.h"
#include "screen_playlist.h"
#include "screen_playing.h"
#include "screen_tracklist.h"
#include "mpd.h"

/* Returns TRUE iff the semaphore count has reached zero, i. e. the semaphore is available */
#define PT_SEM_CHECK(pt, s) (0 == (s)->count) 


// Length of response line
#define RESPLEN 1024

static char response[RESPLEN];

/* Used by the information collecting routines */
static char track_artist[TL_ENTRY_LENGTH];
static char track_title[TL_ENTRY_LENGTH];
static int track_pos;

static int volume_arg;
static enum PLAYSTATE state_arg;
static int song_arg;

/* ------------------------------- Our MODEL ----------------------------------------------

	mpd has a potentially very large current playlist.
	Our model knows only about a small part of this playlist.
	To simplify things we always start with the current song of mpd as our first item in the playlist.
	Later we can collect information about the previous 2 or 3 songs, so that the user can scroll backwards fast.
	From this current song onward we know about around the 10 or so next songs.
	(Our view may indeed show only a small portion of this information, but the user may scroll therein.)

	We record the mpd playlist position of the current song in the variable offset_from_mpd.
	The next n songs in the playlist fill the remaining slots of our model.

	The play_list contains title and artist of the current playlist
	play_list[0] is information for the currently playing song.
	play_list[1] .. play_list[n] are information for the next n songs in the playlist 
*/

// We define a function model_has_changed() in mpd.h to tell us if the model ( i.e. playlist) has changed
// A second function reset_model_change() resets the changed flag.


/* Set by function track_list_needed() and read by handle_user_req() */
static int req_track_no;


/* ------------------------------ The Controller ----------------------------------------- */

/* This semaphore protects access to communication with mpd. 
	It is initialized to 1, so that only 1 thread at a time can communicate with mpd.
	Communication means sending a command and receiving the answer.
	We must protect it, so that only 1 thread is potentially interested in the answer from mpd.
	And so that no other thread can send a command before an answer has arrived (or a timeout).
*/
static struct pt_sem mpd_mutex;


/* Variable which is set by the response collecting routines.
	When we got a response with "OK" we set this to 1.
	If no response or "ACK" then it will stay at 0.
	TODO could as well be implemented as a semaphore
*/
static int response_valid;

/* We got a line from mpd after a current_song command.
	Gather the information in the variable mpd_status.
	Sets response_valid to 1 if we received an "OK" string.
*/ 
static void 
process_currentsong_line(char *s){

	/* Compare with "Artist: " */
	if (strstart(response, "Artist: "))
		mpd_set_artist(response+8);
			
	/* Compare with "Title: " */
	if (strstart(response, "Title: "))
		mpd_set_title(response+7);
	
	/* Compare with "Pos: " */
	if (strstart(response, "Pos: ")){
		mpd_set_pos (atoi (response+5));
	};
#if 0	
	/* Compare with "Id: " */
	/* We store this info so that we can detect if we have any current song at all */	
	if (strstart(response, "Id: ")){
		tmp = atoi (response+4);
		if (tmp != mpd_status.songid){
			/* We really have a new song. */
			mpd_status.songid = tmp;
			mpd_status.time_elapsed = -1;
			mpd_status.time_total = -1;	
			model_changed(TITLE_CHANGED);
			model_changed(ARTIST_CHANGED);
			model_changed(TIME_CHANGED);
		};
	};
#endif	
	/* Compare with "OK" */
	if (strstart(response, "OK"))
		response_valid = 1;
};

// NOTE this is a bit tricky. We can get multiple lines which start with "playlist:"
//		We store them consecutively in an array. So we must make sure to reset the index 
//		into the array to 0 before using process_lsinfo_line !
//		This is done in handle_cmd() and not very elegantly solved, but worst case is the playlist is not reset but expanded,
//		so that all playlist names are doubled. Not dramatic.
static void 
process_lsinfo_line(char *s){
	
	/* Compare with "playlist: " */
	if (strstart(response, "playlist: "))
		mpd_set_playlistname(response+10);

	/* Compare with "OK" */
	if (strstart(response, "OK")) {
		response_valid = 1;
	};
};

/* We got a line from mpd after a status command.
	Gather the information in the variable mpd_status.
	Sets response_valid to 1 if we received an "OK" string.
*/ 
void 
process_status_line(char *s){
	int tmp;
	char *s2;
	
	if (strstart(response, "playlistlength: ")) 
		set_playlistlength(atoi(response+16));
		
	if (strstart(response, "volume: "))
		mpd_set_volume (atoi(response+8));
	
	if (strstart(response, "state: ")){
		tmp = UNKNOWN;
		if (strstart(response+7, "play")) 
			tmp = PLAY;
		else 
			if (strstart(response+7, "pause"))
				tmp = PAUSE;
			else
				if (strstart(response+7, "stop"))
					tmp = STOP;
		mpd_set_state(tmp);
	};
	
	/* Compare with "song: " */
	if (strstart(response, "song: ")) 
			mpd_set_pos (atoi(response+6));
	
	/* TODO maybe read songid as well */
	
	if (strstart(response, "time: ")) {
		s2 = strchr(response+6, ':');
		if (s2)
			mpd_set_time( atoi(response+6), atoi(s2+1) );
	};
	
	/* Compare with "OK" */
	if (strstart(response, "OK")) {
		response_valid = 1;
	};
};


/* We got a line from mpd and expect only OK or ACK answer.
	Sets response_valid to 1 if we received an "OK" string.
*/ 
void 
process_ok(char *s){
	/* Compare with "OK" */
	if (strstart(response, "OK")) {
		response_valid = 1;
	};
};

/* We sent a "CLEAR" command
	Set mpd_status.cur_playlist to empty and response_valid to 1 if we received an "OK" string.
*/ 
void 
process_clear_line(char *s){
	/* Compare with "OK" */
	if (strstart(s, "OK")) {
		/* A lot of things happen when a clear command has been executed: */
		
		/* The length of the current playlist is 0. 
			Adjust the tracklist if necessary 
		*/
		set_playlistlength(0);
		
		/* MPD stops playing */
		mpd_set_state(STOP);
		
		response_valid = 1;
	};
};

/* We sent a "LOAD" command
	Set mpd_status.cur_playlist to user_status.cur_playlist and response_valid to 1 if we received an "OK" string.
*/ 
void 
process_load_line(char *s){
	if (strstart(s, "OK")) {
		mpd_set_newplaylist();
		response_valid = 1;
	}
};

/*
	The information gathered from playlistinfo response is stored in special variables.
	These must have been reset to the correct values before hand.
*/
void 
process_playlistinfo_line(char *s){
	
	/* Compare with "Artist: " */
	if (strstart(response, "Artist: ")){
		strn_cpy(track_artist, response+8, TL_ENTRY_LENGTH-1);
	};
			
	/* Compare with "Title: " */
	if (strstart(response, "Title: ")){
		strn_cpy(track_title, response+7, TL_ENTRY_LENGTH-1);
	};

	/* Compare with "Pos: " */
	if (strstart(response, "Pos: ")) {
		track_pos = atoi(response+5);
	};

	/* Compare with "ACK" */
	/* It happens not too infrequently, that mpd can not find a song in the tracklist (database changed).
		We could issue a delete command or inform the user if this is the case.
		For now we just set the info to a value of "- not - found -" so that we do not repeatedly ask mpd about it.
		We could also detect tracks without title and name tags and substitute the filename instead.
	*/
	if (strstart(response, "ACK")) {
		strn_cpy(track_artist, "- not", TL_ENTRY_LENGTH-1);
		strn_cpy(track_title, "found -", TL_ENTRY_LENGTH-1);
		if (track_pos == req_track_no)
			tl_store_track(track_title, track_artist, track_pos);
	};
	
	/* Compare with "OK" */
		//We could here detect tracks without title and name tags and substitute the filename instead.
	if (strstart(response, "OK")) {
		if (track_pos == req_track_no)
			tl_store_track(track_title, track_artist, track_pos);
		response_valid = 1;
	};
		
};

/* We sent a "setvol" command */
static void 
process_volume_line(char *s){
	if (strstart(s, "OK")) {
		mpd_set_volume(volume_arg);	
		response_valid = 1;
	};
};

/* We sent a state changing command (PLAY, PAUSE, STOP) */
static void 
process_state_line(char *s){
	if (strstart(s, "OK")) {
		mpd_set_state(state_arg);	
		response_valid = 1;
	};
};

/* We sent a "PLAY xxx" command. */
static void 
process_play_line(char *s){
	if (strstart(s, "OK")) {
		mpd_set_state(PLAY);
		mpd_set_pos(song_arg);
		response_valid = 1;
	};
};

/* We sent a "SEEK pos xxx" command. */
static void 
process_seek_line(char *s){
	if (strstart(s, "OK")) {
		if (mpd_get_state() == STOP)
			mpd_set_state(PLAY);		// seeking in MPD automatically leaves STOP state and enters PLAY state, sad but true
		mpd_set_time(-1,-1);		// better reread the exact time
		response_valid = 1;
	};
};



/* This semaphore is <> 0 iff a response line from mpd is ready. */
static struct pt_sem line_ready; 


/*  Whenever the radio receives a packet, it is handed to this routine.
	We maintain a buffer (response[] and response_ptr) to store partial payload data.
	In theory, multiple packets should be assembled together to get the full mpd response into the buffer.
	Then we should parse the buffer contents.
	BUT: The data sent by mpd can be very large (for example a complete playlist).
	There are three possible ways around this situation:
	- Only issue commands which we know generate small enough responses.
	- Filter data by sending program mpdtool on the mpd server. 
		We use this approach with the LSINFO command because else we get overwhelmed by responses.
	- Process parts of the response, even when it is not complete and discard them. This works, because 
	  mpd sends responses in the form of small lines with keyword : value pairs and seperated by '\n'. 
	
	We have chosen the last approach here.
	The assemble_line() thread assembles payload packets until a complete line has arrived.
	The current line is stored in the variable response[] as a null-terminated string without the newline.
	When one complete line has been assembled, the semaphore line_ready is signaled.
	Then we wait until the line has been processed and line_ready is free again 
	or until a timeout occurs. 

	Thread does never end.
	
	TODO this could be made more efficient. We do not need an extra response buffer.
	The rf module increments a semaphore each time it puts a complete line in its buffer.
	The consuming function(s) (collect_lines) decrement that semaphore.
	The buffer contents need not be copied if the rf module substitutes each "\n" with a '\0' to make C strings.
	But we have to make sure that lines are consumed or discarded within a relatively short time frame, so that the buffer
	is not overflowing.
*/

PT_THREAD (assemble_line(struct pt *pt)) {
	static uint8_t c;
	static int response_ptr = 0;
	static struct timer tmr;
	
	PT_BEGIN(pt);
	while (1){
		// Block until we can get data from the receive-buffer.
		PT_WAIT_WHILE(pt, rx_buf_empty());
		
		/* Copy data to response buffer until we reach '\n' */
		c = get_from_rx_buf();

		if (c != '\n'){
			if (response_ptr < (RESPLEN - 1))	// if the line is too long, ignore superfluous characters.
				response[response_ptr++] = c;		
		} else {	
			response[response_ptr] = 0;		// Make it a valid C string

			/* Signal a semaphore that there is a line in response[]. */
			PT_SEM_SIGNAL(pt, &line_ready);
			
			debug_out(response, 0);
			
			/* Wait until the response buffer has been processed 
				If no one is interested in the line, we forget it.
			*/
			timer_add(&tmr, 5*TICKS_PER_TENTH_SEC, 0);
			PT_WAIT_UNTIL(pt, PT_SEM_CHECK(pt, &line_ready) || timer_expired(&tmr));
			timer_del(&tmr);
	
			/* Either way, reset the semaphore */
			PT_SEM_INIT(&line_ready, 0);

			response_ptr = 0;	
		};
	};
	PT_END(pt);
};


/* We have at least one line of response to our command from mpd.
	Collect this and the following lines until "OK" or "ACK" of time out.
	Each line collected is processed with the function process_line.
	Ends with valid_response set to 1 iff we finally received an OK.
	If ACK or no answer within some time delay exit with response_valid set to 0.
*/
PT_THREAD (collect_lines(struct pt *pt, void (*process_line) (char *s))  ){
	static struct timer tmr;
	
	PT_BEGIN(pt);
	/* We should receive all the answers in a relatively short time frame, else something went wrong anyway */
	timer_add(&tmr, 1*TICKS_PER_SEC, 0);
	response_valid = 0;
	
	do { 
		/* gather information from response line by the given function */
		process_line(response);
		PT_SEM_INIT(&line_ready, 0);

		if (response_valid){
			debug_out("valid answer ", 0);
			break;		/* and end the thread */
		};	
						
		// Wait for next line
		PT_WAIT_UNTIL(pt, (!PT_SEM_CHECK(pt, &line_ready) || timer_expired(&tmr)) ); 
	
	} while (!response_valid && !timer_expired(&tmr));
	
	timer_del(&tmr);	
	PT_END(pt);
};



/* Thread to send a command, collect replies from mpd and do error handling 
	The command string sent to mpd is given as second argument.
	Third argument is a function which is called to collect responses.
	Fourth argument is a function which is called after we got a valid response ("OK"). // NOTE not used !
	Fifth argument tells us how often we should try to send the command.
	TODO we could add more parameters concerning time out and error handling.
	Thread waits for semaphore command_empty.
*/
PT_THREAD (handle_cmd(struct pt *pt, char *cmd, void (*process_line) (char *s), void (*process_info) (void), int max_tries)){
	static struct timer tmr;
	static int tries;
	static struct pt child_pt;
	
	PT_BEGIN(pt);
	
	/* We do not send immediately if we are above our error threshold. */
	if (get_comm_error() > 14){
		timer_add(&tmr, 10*TICKS_PER_TENTH_SEC, 0);
		PT_WAIT_UNTIL(pt,  timer_expired(&tmr));
		timer_del(&tmr);
	};
	
	/* We want to acquire access to mpd for this command and its answer */
	PT_SEM_WAIT(pt, &mpd_mutex);

	for (tries = 0; tries < max_tries; tries++){	
		
		// TODO after acquiring exclusive access to mpd initialize line_ready ?
		//		we don't want old answers to fool this routine 
		
		debug_out(cmd, 0);
		send_cmd(cmd);
	
		/* Time that we want to wait for first response line */
		timer_add(&tmr, 5*TICKS_PER_TENTH_SEC, 0);
		
		PT_WAIT_UNTIL(pt, !PT_SEM_CHECK(pt, &line_ready) || timer_expired(&tmr));
		
		if (timer_expired(&tmr)){
			error_inc(3);
			timer_del(&tmr);
			debug_out("retry ", tries);
			continue;
		};
	
		timer_del(&tmr);
		if (! PT_SEM_CHECK(pt, &line_ready) ){
			
			/* This thread collects all information from mpd and sets response_valid. */
			PT_SPAWN(pt, &child_pt, collect_lines(&child_pt, process_line));

			if (response_valid){
				error_dec(1);
				/* Now our cmd got an OK. Process the info gathered (if necessary) */
				if (process_info) process_info();
				break;
			} else
				error_inc(1);
				
		} else 
			error_inc(3);
	};
	
	/* Now other threads can again communicate with mpd */
	PT_SEM_SIGNAL(pt, &mpd_mutex);
		
	if (tries >= max_tries) 
			debug_out("No answer ", 1);
		// mpd did not respond. Some clever error handling here	TODO maybe really do some error handling ?	
	
	PT_END(pt);
};
	

/* This routine is called when our model has changed in some way. 
	We inform the view about the changed items.
	TODO better make this a seperate task, so that we do not spend too much time in 
		this routine 
*/
void 
inform_view(int model_changed){
	if (model_changed & VOLUME_CHANGED)
		view_volume_changed(mpd_get_volume());
	
	if (model_changed & TITLE_CHANGED){
		if (mpd_get_curtitle() == NULL)
			view_title_changed("");
		else	
			view_title_changed(mpd_get_curtitle());
	};
	if (model_changed & ARTIST_CHANGED){
		if (mpd_get_curartist() == NULL)
			view_artist_changed("");
		else
			view_artist_changed(mpd_get_curartist());
	};
	if (model_changed & TIME_CHANGED)
		view_time_changed(mpd_get_time_elapsed(), mpd_get_time_total());
	
	if (model_changed & STATE_CHANGED)
		view_state_changed(mpd_get_state());
	
	if (model_changed & POS_CHANGED)
		view_pos_changed(mpd_get_song());
	
	if (model_changed & TRACKLIST_CHANGED)
		view_tracklist_changed();

	if (model_changed & PL_NAMES_CHANGED){
		view_playlist_changed();
	};

	model_reset_changed();
};
	

/* This thread keeps record of the remaining time of a currently playing song. 
	We could issue a status command every second and get the playing time exactly as seen by mpd.
	But this generates quite some traffic over the air, which we want to avoid.
	So we update the time seen by the user here until we reach the end of the currently playing song.
	After the song has ended, we tell the controller that we need a new status from mpd to get the
	correct time.
*/
PT_THREAD (update_playtime(struct pt *pt)){
	static struct timer sec_tmr;
	
	PT_BEGIN(pt);
	
	/* We want to be noticed once per second. */
	timer_add(&sec_tmr, 1*TICKS_PER_SEC, 1*TICKS_PER_SEC);
	
	while (1){
		PT_WAIT_UNTIL(pt, timer_expired(&sec_tmr));
		sec_tmr.expired = 0;
		mpd_inc_time();
	};
	PT_END(pt);
};

/* This routine checks for direct input from user (via keyboard) 
	It calls user_request() and if necessary user_request_arg()
	This routine reflects all of the user wishes in the model,
	so that the necessary action can later be detected and acted upon.
*/
static void 
handle_user_wish(void){
	enum USER_CMD cmd = user_request();
	
	switch (cmd) {
		case PLAY_CMD:
			user_set_state(PLAY);
			break;
			
		case STOP_CMD:
			user_set_state(STOP);
			break;
			
		case PAUSE_TOGGLE:	
			user_toggle_pause();
			break;
	
		case VOLUME_UP:
			mpd_want_volume_add(+5);
			break;

		case VOLUME_DOWN:
			mpd_want_volume_add(-5);
			break;

		case MUTE_CMD:
			user_mute();
			break;
	
		case SEL_SONG:
			mpd_want_song(user_request_arg());
			break;
			
		case PREV_SONG:
			mpd_want_song_add(-1);	
			break;
			
		case NEXT_SONG:
			mpd_want_song_add(+1);	
			break;
				
		case FORWARD_CMD:
			mpd_want_time_add(+10);	
			break;
			
		case REWIND_CMD:
			mpd_want_time_add(-10);	
			break;

		case LOAD_CMD:
			mpd_want_playlist(user_request_arg());
			break;

		default:
			break;
	};
};



/* This function returns the type of action, if the controller has to initiate some action,
	i.e. some commands have to be sent to mpd.
	In effect, this routine checks if user model and mpd model disagree.
	If they disagree, we return the needed action.
	It also checks for direct input from user.	
	If the action needs an argument, it is stored in *arg.
	
*/
static enum USER_CMD
action_needed(int *arg){

	/* We query the user for action and reflect changes in the model. */
	handle_user_wish();
	
	/* We check the model if any action is necessary */
	return model_needs_action(arg);
}


#define CMDSTR_LEN 100	
/* Some action (communication with mpd) is needed. Start this. */
static
PT_THREAD (exec_action(struct pt *pt, enum USER_CMD cmd, int arg) ){
	static struct pt child_pt;
	static char cmd_str[CMDSTR_LEN];
	
	PT_BEGIN(pt);
	
	/* Here we can check for communication errors.
		If we have too many errors, we try to solve them by resetting our
		radio reception to a sane state if necessary.
	*/
	if (get_comm_error() > 10){
		rx_reset();
	};
			
	/* Volume changing command */
	if ( cmd == VOLUME_NEW ) {
		compose_string (cmd_str, "setvol ", arg);
		volume_arg = arg;
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, cmd_str, process_volume_line, NULL, 1));
		PT_EXIT(pt);
	};

	if (cmd == NEXT_SONG){
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "next\n", process_ok, NULL, 1));
		/* We do not really know title and artist of the next song. There may not even be one.
			We could guess the info from our tracklist, but safer to reread the info later */
		mpd_set_pos(SONG_UNKNOWN);
		/* The user should not care what the next song is, he just wanted the next one. */
		/* We set the user expectation to unknown, so he accepts any further changes */	
		user_song_unknown();
		PT_EXIT(pt);
	};
	
	if (cmd == PREV_SONG){
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "previous\n", process_ok, NULL, 1));
		mpd_set_pos(SONG_UNKNOWN);
		user_song_unknown();
		PT_EXIT(pt);
	};
	
	if (cmd == SEL_SONG){
		compose_string (cmd_str, "play ", arg);
		song_arg = arg;
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, cmd_str, process_play_line, NULL, 1));
		PT_EXIT(pt);
	};
	
	if (cmd == PLAY_CMD){
		state_arg = PLAY;
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "play\n", process_state_line, NULL, 1));
		PT_EXIT(pt);
	};
	
	if (cmd == STOP_CMD){
		state_arg = STOP;
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "stop\n", process_state_line, NULL, 1));
		PT_EXIT(pt);
	};
	
	if (cmd == PAUSE_ON){		
		state_arg = PAUSE;
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "pause 1\n", process_state_line, NULL, 1));
		PT_EXIT(pt);
	};
	
	if (cmd == PAUSE_OFF){		
		state_arg = PLAY;
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "pause 0\n", process_state_line, NULL, 1));
		PT_EXIT(pt);
	};
		
	/* The seek command only works if we have a current song to give to mpd (why that?) */		
	if ( ( (cmd == FORWARD_CMD) || (cmd == REWIND_CMD) ) && (mpd_get_song() >= 0) ){
		compose_string2 (cmd_str, "seek ", mpd_get_song(), arg);
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, cmd_str, process_seek_line, NULL, 1));
		PT_EXIT(pt);
	};

	if (cmd == CUR_SONG_CMD){
		/* We set these values just in case that the current track has no title or artist */
		mpd_set_artist("???");
		mpd_set_title("???");
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "currentsong\n", process_currentsong_line, NULL, 1));
		PT_EXIT(pt);
	};
	
	if (cmd == STATUS_CMD){
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "status\n", process_status_line, NULL, 1));	
		PT_EXIT(pt);
	};	
	
	if (cmd == PLINFO_CMD){
		compose_string (cmd_str, "playlistinfo ", arg);
		track_pos = -1;
		track_title[0] = '\0';
		track_artist[0] = '\0';
		req_track_no = arg;
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, cmd_str, process_playlistinfo_line, NULL, 1));	
		PT_EXIT(pt);
	};	
	
	if (cmd == LSINFO_CMD){
		start_reading_playlists();
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "lsinfo\n", process_lsinfo_line, NULL, 1));	
		PT_EXIT(pt);
	};	
	
	if (cmd == CLEAR_CMD){
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, "clear\n", process_clear_line, NULL, 1));	
		PT_EXIT(pt);
	};	

	if (cmd == LOAD_CMD){
		strn_cpy (cmd_str, "load \"", CMDSTR_LEN);
		strn_cat (cmd_str, pl_get_name(arg), CMDSTR_LEN);
		strn_cat (cmd_str, "\"\n", CMDSTR_LEN);
		PT_SPAWN(pt, &child_pt, handle_cmd(&child_pt, cmd_str, process_load_line, NULL, 1));	
		PT_EXIT(pt);
	};	
	
	/* We do not know this command. */
	debug_out("Unknow Request ",cmd);

	PT_END(pt);
};


/* How does the controller work ? 
	We have a model of the status of mpd.
	We assume this is ok until we know otherwise.
	The status of mpd is reflected in the view.
	
	Now when the user is not satisfied with the current state, he initiates some action.
	We reflect the changes that the user wants in the model.
	
	Now whenever the user status and the mpd status differ, 
	the controller issues one ore more commands to mpd to get to the user desired status.
	It also sets the potentially changing items in mpd_status to unknown, because we
	do not know for sure if and how mpd reacts to the command.
	Sometimes we get the new mpd status directly with the answer, so we are done.
	Sometimes mpd does not react or does not tell us enough details in its answer.
	So:
	The controller continuously checks for unknown status items in mpd_status.
	It then sends commands to get the unknown items and updates mpd_status.
	This way we repeatedly try commands until user_status and mpd_status are compatible.
	
	A complication:
	We have a status item that is frequently changing without user intervention, namely the
	remainig playing time of the current song. It would be inefficient and consume wireless bandwidth 
	to check every second for the remaining playing time, because it is predictable.
	We use a seperate task to update the time information regularily. We also assume 
	that the user is expecting the passage of time and set the same time in the user model.
	
	Further complications:
	mpd might change its own status expectedly and unexpectedly.
	mpd changes status expectedly whenever the current song has ended.
	It might also change status unexpectedly due to an error.

	We must also reflect the user expectation when listening to a playlist, that after a song has ended
	the next song will be played. We can do that by setting song and songid in user_status to unknown, so 
	that the user is accepting any change in these items.
	
	NOTE  We should build in a failure mode, where we check for error messages from mpd and
	whether we can reach mpd at all.
*/

/* ### Controller Task ###
	Started once. Never returns.
	
	Starts a thread to handle sending commands to mpd (handle_cmd).
	Starts a thread to assemble a complete line from mpd rf ring buffer (assemble_line).
	
	Main loop:
	Waits for a necessary action or a change in the model
	and executes the action and informs the view accordingly.
	
*/
PT_THREAD (controller(struct pt *pt)){
	static struct pt action_pt;
	static enum USER_CMD action;
	static int action_arg;
	
	PT_BEGIN(pt);
		
	/* Make sure 1 and only 1 available thread can communicate with mpd */
	PT_SEM_INIT(&mpd_mutex, 1);

	/* At start of program we should have no response line from mpd */
	PT_SEM_INIT(&line_ready, 0);
	task_add(&assemble_line);
	
	task_add(&update_playtime);	
	
	while (1) {
		/* This is a bit tricky. We must make sure action is set and is called only once. So it is at the beginning */
		PT_WAIT_UNTIL(pt, ( (action = action_needed(&action_arg)) != NO_CMD) || model_get_changed() );
				
		if ( action != NO_CMD ){
			debug_out("action ", action);
			PT_SPAWN( pt, &action_pt, exec_action(&action_pt, action, action_arg) );
			PT_YIELD(pt);
		};
		
		if ( model_get_changed() ) {
			inform_view(model_get_changed());
			PT_YIELD(pt);
		};
	};
	PT_END(pt);
};


/* ------------------------------ End of Controller ----------------------------------------- */
