/*
    main.c
    Copyright (C) 2007  

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <P89LPC932.h>               /* special function register declarations   */
                                    /* for the Philips P89C931 device           */

#include "cc1100.h"
#include "smartrf_CC1100.h"
#include "serial.h"
#include "cmd.h"


#define EOT	0x04
#define ETX	0x03
#define ACK	0x06
#define ESC	0x1B

// Software Reset bit of AUXR1
#define SRST	3

// Bits of RTCCON
#define RTCF	7
#define RTCS1	6
#define RTCS0	5
#define RTCEN	0

// Bits of SSTAT
#define CIDIS	5

// Bits of IEN0
//#define ESR	4
//#define EA	7

// A simple random number generator, TODO debug only
#define znew  ((z=36969*(z&65535)+(z>>16))<<16)
#define wnew  ((w=18000*(w&65535)+(w>>16))&65535)
#define RAND  (znew+wnew)

static unsigned long z=362436069, w=521288629;
void setseed(unsigned long i1,unsigned long i2){z=i1; w=i2;}


/* The buffer is a bit longer than MAX_PKTLEN, so that we can buffer a few bytes after sending XOFF 
	It is implemented as a FIFO ring buffer
*/
#define BUFSIZE 80
idata char buf[BUFSIZE];
unsigned char bufstart;
unsigned char pkt_len=0;

volatile unsigned char bufcnt;
volatile unsigned char buflim;
volatile bit got_etx;
volatile bit got_eot;
volatile unsigned char dropped;

/* This byte tells us if we are sending a cmd (mode = 0) or getting a response (mode = 1) */
unsigned char mode;
#define cmd_mode_start 	0
#define cmd_mode 	1
#define resp_mode_start 2
#define resp_mode	3
#define resp_mode_finish 4
#define waittx_mode 	5

#define send_esc(s) {send_byte(ESC); send_string(s); send_byte(ESC);} 


/* This byte is the status of the handle_tx routine */
unsigned char tx_state;

#define TX_IDLE		0
#define TX_COPY		1
#define TX_FILL		2
#define TX_PRESEND	3
#define	TX_POSTSEND	4

/* Number of payload bytes in packet (not counting length and address and suffix bytes) */
#define PKTLEN 60

unsigned short total_len;

unsigned char timeout;

unsigned char cc1100Addr;
idata unsigned char cnt[4]; 

__sfr __at (0xAD) CMP2_   ;


unsigned char min(unsigned char x, unsigned char y){
	if (x <= y) return x;
	return y;
}


/* Our buffer has the following semantics:
	- We put data bytes in sequentially
	- The buffer is organizing them automatically into packets suitable for radio transmission
	- buf[bufstart] is 0, if no data payload bytes are in buffer, i.e. buffer empty
	- buf[bufstart] is n with n<0, if n payload bytes are in buffer, but packet is not complete yet
	- buf[bufstart] is n (with n>0), if complete packet with n payload bytes is in buffer
	- buf[bufstart] is only valid when we are not currently transferring bytes out of the buffer to TXFIFO
*/
void buffer_init(){
	bufcnt = 0;	// Number of bytes in the buffer
	buflim = 0;	// Index of next free place in buffer
	bufstart = 0;	// Index of first data byte in buffer
}

/* "First in" part of buffer routines is handled by ISR: */
/* The serial interrupt reads a character from serial interface if there is one.
	The character is stored in the ring buffer.
	If the ring buffer is full, the last character will be overwritten.
	This ensures that EOT character should be stored.
	The flag got_etx is set to 1 if an ETX character is received.
	ETX will never be stored in buffer.
	For debugging purposes the number of overwritten(=dropped) bytes is counted.
	The flag got_etx can be used as a semaphore, sdcc will later generate atomical TestAndClear instructions for it.
	The flag got_eot is set as soon as we see an EOT character. All following bytes will be dropped until got_eot is cleared.
	bufcnt and buflim (and got_etx and got_eot and dropped) are changed in this routine.
*/

void serial_isr (void) __interrupt (4) {
	char x;
	
	x=SBUF;
	RI = 0;
	
	if (x == ETX){
		got_etx = 1;
		return;
	};
			
	if (got_eot){
		dropped++;
		return;
	};
	

	/* If the buffer is full, overwrite previous byte */
	if (bufcnt >= BUFSIZE){
		dropped++;
		bufcnt--;
		if (buflim == 0)
			buflim = BUFSIZE-1;
		else
			buflim--;
	};
	
	/* Store byte */
	buf[buflim++] = x;
	if (buflim >= BUFSIZE)
		buflim = 0;
	bufcnt++;
	
	if (x == EOT)
		got_eot = 1;

	return;
}


/* "first out" part of buffer routine 
	Returns 0 if the buffer is empty
*/
char buffer_out(){
	char x;
	
	if (bufcnt == 0) return (0);
	x = buf[bufstart++];
	if (bufstart >= BUFSIZE)
		bufstart = 0;
	
	/* The following works only if decrement is an atomic operation */
	bufcnt--;
	
	return x;
}


/* Returns 1, iff there are some empty places in the buffer */
unsigned char has_room(){
	return ( (BUFSIZE - bufcnt) > 18 );
}


/* Returns 1 if we are currently transferring bytes to TXFIFO 
	or sending the TXFIFO over radio
*/
unsigned char tx_busy(){
	/* Is radio idle */
	return (tx_state != TX_IDLE);
}


/* Start radio reception */
void start_rx() {
	// Reset state and set radio in RX mode
	// Safe to set states, as radio is IDLE
	switch_to_idle();
	cc1100_strobe(SFRX);	
	cc1100_strobe(SCAL);
	// TODO probably not necessary, the SRX strobe waits until CC1100 ready
//	while ( (cc1100_read_status_reg_otf(MARCSTATE) & 0x1f) != 0x01) {}; 
	cc1100_strobe(SRX);
}


/*
	Send buffer contents over radio.
	The buffer is telling us if it is ready (i.e at least one packet complete).
	The buffer also tells us how long the packet is.
	We then wait until radio is idle.
	Then we transfer the bytes from buffer to TXFIFO.
	If packet is completely in TXFIFO we start transmission.
	
	This is a state machine. Transmission can be in 5 states:
	(0) TX_IDLE		; not transmitting and nothing to transmit
	(1) TX_COPY		; copy bytes from buffer to tx_fifo
	(2) TX_FILL		; fill the last packet with null bytes
	(3) TX_PRESEND		; waiting for cc1100 ready to send
	(4) TX_POSTSEND		; waiting for cc1100 to finish sending
	
	This routine returns 1 iff the last packet (the one with EOT) has been transferred to TXFIFO (but not sent yet)
	else it returns 0.
*/
unsigned char handle_tx(){
	char x;
	unsigned char ret = 0;
	
	switch (tx_state){
		
	case TX_IDLE:
		/* Is a new packet ready ? */
		if (got_eot || (bufcnt>=PKTLEN) ){
			pkt_len = 0;	/* pkt_len = number of payload bytes transferred to TXFIFO (not counting address) */
			
			/* Transfer the length byte to TXFIFO (just add 1 for the address) */
			cc1100_write1(TX_fifo, PKTLEN+1);
			
			/* Transfer the address to TXFIFO */	
			cc1100_write1(TX_fifo, DEV_ADDR);	
			
			tx_state = TX_COPY;
		};
		break;
		
	case TX_COPY:
		/* Are there still bytes to copy to TXFIFO? */
		if (pkt_len < PKTLEN){
			x = buffer_out();
			cc1100_write1(TX_fifo, x);	
			pkt_len++;
			total_len++;
			if (x == EOT){
				got_eot = 0;
				tx_state = TX_FILL;
				ret = 1;
			};
		} else {
			/* Finished copying ! */ 
			tx_state = TX_PRESEND;
		};
		break;
		
	case TX_FILL:
		if (pkt_len < PKTLEN){
			cc1100_write1(TX_fifo, 0);
			pkt_len++;
		} else
			/* Finished copying ! */ 
			tx_state = TX_PRESEND;	
		break;
		
	case TX_PRESEND:
		/* Are we ready to initiate sending ? */
		if ((cc1100_read_status_reg_otf(MARCSTATE) & 0x1f) == MARCSTATE_IDLE){
			cc1100_strobe(SCAL);
			cc1100_strobe(STX);
			tx_state = TX_POSTSEND;
		};
		break;
	
	case TX_POSTSEND:
		/* Are we finished sending ? */
		if ((cc1100_read_status_reg_otf(MARCSTATE) & 0x1f) == MARCSTATE_IDLE){
			// TODO we can check here for txfifo == 0 to make sure we sent the data
			tx_state = TX_IDLE;
//			send_byte(ESC);send_string("transferred \n");send_byte(ESC);
		};
		break;		
	}
	return (ret);
}

void main(void) {

        unsigned char x;
	unsigned char i;
	cnt[3]=0;
	cnt[2]=0;
	cnt[1]=0;
	cnt[0]=0;
	
	/* -------------------------- Initialize the ports ------------------------------ */
	// All pins are input only after reset, i.e. PxM1=11111111 and PxM2=00000000
	
	// Port0 
	// Push-Pull: Pin 7, 4, 1
	// bidirectional: Pin 0
	// Input only: Pin 6, 5, 3, 2
        P0M1 &= 0x6c;		// 0110 1100 -> P0M1
        P0M2 |= 0x92;		// 1001 0010 -> P0M2
	
	// Port0
        P0 = 0x83;		// 1000 0011 -> Port0
        
	// Port1 Ouput Mode 1	
        P1M1 &= 0x3e;
	
	// Port1 Ouput Mode 2
        P1M2 |= 0xcd;
	
	// Port1
        P1 = 0xcd;
	
	// Port3 Ouput Mode 1
        P3M1 &= 0xfe;
	// Port3 Ouput Mode 2
	P3M2 |= 0x01;
	
	// Port3
        P3 = 0;
	/* ----------------------------------------------------------------- */
	
	
        /*
	PT0AD = 0x24;
        TMOD = 0x22;
        TAMOD = 0x10;
        TCON = 0x50;
        AUXR1 = 0x20;
        TH0 = 0x22;
        TH1 = 0x80;
        CMP2_ = 0x24;
        
        KBPATN = 0x08;
        KBMASK = 0x08;
        KBCON = 0x02;
	*/
	
	/* Enable Break detection and enter ISP when Break occurs */
	AUXR1 |= (1<<6);
	
	RTCCON |= 1<<RTCS1;
	RTCCON |= 1<<RTCS0;
        RTCH = 0xff;
	RTCL = 0xff;
	
        initSerial(384);	// Serial baudrate 38400
	SSTAT |= (1<<CIDIS);	// Combined interrupt disabled, RX and TX generate different interrupts
        
	cc1100_init();

	ESR = 1;
	EA = 1;
	
        cc1100_strobe(SRX);	// Enable receive mode 

	x = cc1100_read_status_reg_otf(CC1100_VERSION); 
//	send_byte(ESC);send_string("CC1100 Version "); send_hex(x); send_string("\n");send_byte(ESC);
//	send_byte(ESC);send_string("Ready\n");send_byte(ESC);

 
	/* Start cmd_mode */
	mode = cmd_mode_start;
	got_etx = 0;
	got_eot = 0;
	
	buffer_init();
	
//	send_byte(ESC);send_string("PATABLE = "); send_hex(cc1100_read1(PATABLE)); send_string("\n");send_byte(ESC);
	

	/* Forever: */
	/* This is a state machine with the variable mode as the state variable */
	while (1) {
		switch (mode) {
		case cmd_mode_start:		// Start receiving over radio, next state is cmd_mode
			RTCCON &= ~((1<<RTCEN) | (1<<RTCF));	// Reset RTC	
			timeout = 100;
			RTCCON |= 1<<RTCEN;		// Enable RTC	
			mode = cmd_mode;
			
			switch_to_idle();

			cc1100_strobe(SFRX);
			cc1100_strobe(SCAL);
			cc1100_strobe(SRX);		// Start receiving
			
			break;

		case cmd_mode:
			/* Wait for input over wireless */
#if 0			
			/* Check if input from serial line */
			if (RI) {
				x = SBUF;
				RI = 0;
				/* We should not receive something here.
					Ignore it, if it is not ESC character
				*/
				if (x == ESC){
					AUXR1 |= (1<<SRST);		// Software Reset !
				};
			};
			
#endif			
			if (GDO0) {			// We have received a complete packet over radio

				cc1100_read(RX_fifo|BURST, cnt, 1);		// Length byte

//				send_byte(ESC); send_string("RCVD: "); send_hex(cnt[0]); send_byte(0x0a); send_byte(ESC); 
				
				cc1100_read(RX_fifo|BURST, cnt+1, 1); 	// Address byte (is subsequently overwritten!)
				
				// We are reading the RX_fifo one byte at a time and sending this byte to serial output 
				for (i=0; i < (cnt[0] - 1); i++){
					cc1100_read(RX_fifo|BURST, cnt+1, 1);
					send_byte(cnt[1]);
				};

				cc1100_strobe(SFRX);	
//				cc1100_getfifo(buf);
				
				/* End of Transmission ? */
//				send_byte(EOT);
				mode = resp_mode_start;
			};
			
		
			if (RTCCON & (1<<RTCF)) {	// Timer overflow
				RTCCON &= ~(1<<RTCF);		// Clear RTC Flag
				if (timeout-- == 0){		// 
					send_byte(EOT);
					
					mode = resp_mode_start;
					break;
				};
			};
			break;	
			
		case resp_mode_start:
			/* We initialize receiving characters from the serial line, next state is resp_mode */
			RTCCON &= ~((1<<RTCEN) | (1<<RTCF));	// Reset RTC	
			timeout = 10;
			RTCCON |= 1<<RTCEN;		// Enable RTC
			mode = resp_mode;
			buffer_init();
			switch_to_idle();
			tx_state = TX_IDLE;
			total_len = 0;
			dropped = 0;
			x = 0;
			break;
			
		case resp_mode:
			/* Response mode, we are waiting for a response via serial line 
				The state stays the same until end of response is detected, then next state is resp_mode_finish
			*/
			
			if (RTCCON & (1<<RTCF)) {	// Timer overflow
				RTCCON &= ~(1<<RTCF);		// Clear RTC Flag
				if (timeout-- == 0)
					mode = resp_mode_finish;
				break;
			};
			
			/* Still enough space in buffer ? */
			if (has_room()){
				/* Is the sender waiting for an ACK ? */
				if (got_etx) {
					got_etx = 0;		// Atomic Operation ! (see sdcc manual)
					send_byte(ACK);
				};
			};
	
			/* Handle transmitting of buffer contents over radio 
				i.e. start sending the buffer contents and check if we have seen the last packet (the one with EOT)
			*/
			if (handle_tx() == 1) {
				/* last packet has started to be sent over radio. */
				/* But the sender may still wait for us to acknowledge his ETX. Do it. */
				if (got_etx) {
					got_etx = 0;		// Atomic Operation ! (see sdcc manual)
					send_byte(ACK);
				};
				
				mode = resp_mode_finish;
				
			};
			break;
				
		case resp_mode_finish:
			/* Finished receiving response. Initialize waiting for end of radio transmission.
				Next mode is waittx_mode. 
			*/
				RTCCON &= ~((1<<RTCEN) | (1<<RTCF));	// Reset RTC	
				timeout = 10;
				RTCCON |= 1<<RTCEN;		// Enable RTC
				mode = waittx_mode;
				break;

		case waittx_mode:
		default:	
			/* must be waittx_mode, wait until tx is finished */
			if (RTCCON & (1<<RTCF)) {	// Timer overflow
				RTCCON &= ~(1<<RTCF);		// Clear RTC Flag
				if (timeout-- == 0){
					mode = cmd_mode_start;
				};
				break;
			};
			
			/* Handle transmitting of buffer contents over radio */
			handle_tx();
		
			/* wait until tx is not busy and then switch to cmd mode*/
			if (tx_busy() == 0)
			{
#if 0				
				send_byte(ESC);send_string("Total bytes sent over radio: "); 
				send_hex(total_len >> 8); send_hex(total_len & 0xff);
				send_string("    Bytes dropped: "); send_hex(dropped);
				send_string("\n");send_byte(ESC);
#endif				
				mode = cmd_mode_start;
				/* If the buffer has room send ACK */
				if (has_room()){
					/* Is the sender waiting for an ACK ? */
					if (got_etx) {
						got_etx = 0;		// Atomic Operation ! (see sdcc manual)
						send_byte(ACK);
					};
				};
			}
			break;
		};	
		
	};

}



