/* Tool to receive mpd commands over serial line and to send back responses */

/*
	Application Layer:
	We do have a half-duplex communication between mpd and its clients.
	All communications are initiated by the clients.
	They send commands to mpd, to which it responds with answers (and actions).
	
	Commands are either single line commands (which end with '\n' (TODO check this))
	or multiple line messages which start with "command_list_begin" and end with "command_list_end\n".
	
	
	MPD signals the end of its (possibly multiple line) answer with a line beginning with either "OK" or "ACK".
	
	
	Link Layer:
	To use the half duplex communication, we need a way to switch between sending and receiving.
	We will use the EOT (End of Transmission) character (0x04). 
	Each party sends this character as the last character when it has finished sending and is ready to receive.

	This program (mpdtool) knows when mpd has finished sending by interpreting the answer. A line starting with "OK"
	or "ACK" signals the last line from mpd's answer.
	
	The betty firmware knows when it has finished sending a command and then sends the EOT character.
	 
	We further need to signal this program (mpdtool) when the serial link is ready to receive the next bunch of characters.
	This program is supposed to run on fast hardware while the serial link hardware might be slow.
	So after sending <MAX_TX> characters, this program sends an EndOfText <ETX> and only starts sending the next bunch when 
	it receives an ACKNOWLEDGE <ACK> from serial line.
	This might not always work. For some as yet unknown reason the serial line firmware does not get the <ETX> or does not send
	the <ACK>. So if we receive other characters while we are waiting for an <ACK> we are assuming an error and stop waiting.
	
	Transport Layer:
	This program receives commands via serial line. It sends them via a TCP/IP socket to mpd.
	The answers are received via the same socket and transferred back over serial line.
	All commands and answers are non-binary characters (ISO-8859-15 I guess).
	
	We must filter some characters: <ETX> <ACK> and <EOT> are not transmitted in either direction.
	If they occur in the input stream, they are simply dropped.

*/

#define VERSION_MAJOR 0
#define VERSION_MINOR 9

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <fcntl.h>
#include <sys/time.h>
#include <time.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>


#define BUFFER_SIZE 1024


#define BAUDRATE B38400

#define _POSIX_SOURCE 1 /* POSIX compliant source */
#define FALSE 0
#define TRUE 1

#define ETX	0x03
#define ACK	0x06
#define EOT	0x04
#define	CR	0x0d
#define LF	0x0a
#define ESC	0x1b

int min(int x, int y){
	if (x <= y)
		return x;
	return y;
};

int write_all(int fileDesc, char * buffer, int bytesToSend){
	int bytesWritten;
	int num = 0;
	
	for (bytesWritten = 0; bytesWritten < bytesToSend; bytesWritten += num)
	{

		/* Write a certain number of bytes to fileDesc. We either send the remaining bytes in buffer
		or the remaining bytes to fill a packet, whichever is shorter.
		*/
		num = write(	fileDesc,	
				(void *)(buffer + bytesWritten),
				 (bytesToSend - bytesWritten)
		);
		/* Did our write fail completely ? */
		if (num == -1) {
			perror("write_all()");
			printf("We only sent %d bytes because of the error!\n", bytesWritten);
			break;
		};
		
		/* Now num is the number of bytes really written. Can be shorter than expected */
	};
	return ( bytesWritten );
};


/*
	Write bytesToSend bytes out on fileDesc
	Handles short writes.
	Keeps a global count of currently written bytes
	and waits for ACK if MAX_TX has been reached. 
	Counts bytes actually written in global cur_tx_len
	returns -1 on failure, 0 on success
*/
#define MAX_TX 16
static int cur_tx_len;

int write_all_ack(int fileDesc, char * buffer, int bytesToSend){
	int bytesWritten;
	int res;
	int num = 0;
	char ETX_char = ETX;
	char in_buffer[1];
	
	for (bytesWritten = 0; bytesWritten < bytesToSend; bytesWritten += num)
	{

		/* Write a certain number of bytes to fileDesc. We either send the remaining bytes in buffer
			or the remaining bytes to fill a packet, whichever is shorter.
		*/
		num = write(	fileDesc,	
			   	(void *)(buffer + bytesWritten),
			 	min( (bytesToSend - bytesWritten), (MAX_TX - cur_tx_len))
		);
		/* Did our write fail completely ? */
		if (num == -1) {
			perror("write_all_ack()");
			printf("We only sent %d bytes because of the error!\n", bytesWritten);
			break;
		};
		
		/* Now num is the number of bytes really written. Can be shorter than expected */
		cur_tx_len += num;

		if (cur_tx_len == MAX_TX){
			res = write(fileDesc, &ETX_char, 1);
			if (res == -1) {
				perror("write_all_ack()");
				printf("We could not write ETX!\n");
				break;
			};
			
			cur_tx_len = 0;
			
			/* Wait until ACK character received or time out */
			in_buffer[0] = 0;
			do {
				res = read(fileDesc, in_buffer, 1);
				if (res == 0) {
					printf("Timeout on ACK! We only sent %d bytes!\n", bytesWritten);
					return(bytesWritten);
				};
				if (in_buffer[0] != ACK) {
					printf ("previous character: <%02x>\n", buffer[bytesWritten + num - 1]);
					printf("Getting char while waiting for ACK <%02x>%c\n",in_buffer[0], in_buffer[0]);	
					printf ("remaining bytes: %d\n", bytesToSend - bytesWritten - num);
				};			
			} while (in_buffer[0] != ACK);

		};		
	};
	return ( bytesWritten );
};


/* 
	Returns 1, if we can read at least 1 character from file descriptor fd
	Else returns 0.
	Uses a time out of 3.5 seconds
*/
int check_read(int read_fd)
{
	struct timeval tv;
	fd_set readfds;
	int res;
	
	tv.tv_sec = 3;
	tv.tv_usec = 500000;

	FD_ZERO(&readfds);
	FD_SET(read_fd, &readfds);

	// don't care about writefds and exceptfds:
	res = select(read_fd+1, &readfds, NULL, NULL, &tv);
	if (res == -1){
		perror("select()");
		return 0;
	};
	
	if (FD_ISSET(read_fd, &readfds)){
		return (1);			// We got a response
	} else
		printf("Timed out.\n");
	
	return 0;
} 


/* Send a data record in intel hex format to device 
	Ceck that it was received correctly.
*/
int send_string(int fd, char *s, char *answer){
	char buffer[BUFFER_SIZE];
	char buffer2[BUFFER_SIZE];
	int cmd_len = strlen(s);
	int i, len;
	
	/* Make sure that all characters are upper case ! */
	for (i=0; i<cmd_len; i++)
		buffer2[i] = toupper(s[i]);
	buffer2[cmd_len] = 0;
	
	write_all(fd, buffer2, cmd_len);
	len = read_boot_response(fd, buffer);
	buffer[len] = 0;
	
	/* Check that our command was echoed correctly */
	if (0 != strncmp(buffer, buffer2, cmd_len)) {	
		printf("Error: Invalid echo!\n");
		printf("%s\n", buffer);
//		return(-20);	
	};
	
	/* buffer[len-1] is 0x0a, buffer[len-2] is 0x0d, buffer[len-3] is '.'  */
	/* The last character should be a '.' */
	if (0 != strncmp(buffer+len-3, ".", 1)) {	
		printf("Error: No . response!\n");
		return (-20);	
	};
	
	/* Print the response characters (if any) */
	for (i=cmd_len; i < (len-3); i++)
		*(answer++) = buffer[i];
	*answer = 0;
	
	return 0;

};
	
/* Read a boot loader response from serial line 
	A response is a single line, terminated by 0x0a
	Returns number of bytes read, if successful
	Returns 0, if 
		- line is longer than BUFFER_SIZE-1
*/
int read_boot_response (int fd, char *buffer){
	int res;		 
	int stop = 0;
	int bytes_read = 0;
	
	do {
		if (bytes_read >= (BUFFER_SIZE - 1)) {
			fprintf(stderr, "Error, too many characters from serial line!\n");
			bytes_read = 0;		// Forget all characters received so far
		};
		
		/* Read at most one character and time out if none received */
		res = read(fd, buffer+bytes_read, 1);
		if (res < 0) continue;
		if (res == 0) {
//			fprintf(stderr, "Time out on serial line\n");
			stop = 1;			// Device has finished sending
		} else {
			stop = (buffer[bytes_read] == 0x0a);
//			printf("Received: %02x\n", (unsigned char) buffer[bytes_read]);
			bytes_read += res;
		};
	} while (!stop);
	
	buffer[bytes_read]=0;             /* set end of string, so we have a real string */

	return (bytes_read);
};

			
/* Read a command from serial line 
	A command is a single line, terminated by '\n'
	Or it is terminated by EOT(0x04)
	Returns number of bytes read (incl. '\n'), if successful
	Returns 0, if time out on serial line
*/
int read_cmd (int fd, char *buffer){
	int res;		 
	int stop = 0;
	int bytes_read = 0;
//	int escaped = 0;
	
	do {
		if (bytes_read >= (BUFFER_SIZE - 1)) {
			fprintf(stderr, "Error, too many characters from serial line!\n");	
			bytes_read = 0;		// Forget all characters received so far
			continue;
		};
		
		/* Read at most one character and time out if none received */
		res = read(fd, buffer+bytes_read, 1);
		
		if (res < 0){
			printf("Error on read from serial: %d\n", errno);
			continue;
		};
		
		if (res == 0) {
			fprintf(stderr, "Time out on serial line\n");
			bytes_read = 0;			// Assume error. Forget all characters received so far
//			escaped = 0;
			continue;
		} ;
#if 0		
		if (escaped){
			if (buffer[bytes_read] == ESC)
				escaped = 0;
			else
				fprintf(stderr,"%c",buffer[bytes_read]);
			continue;
		};
#endif
		switch (buffer[bytes_read]) {
			case EOT:
				stop = 1;
				break;
#if 0				
			case ESC:
				escaped = 1;
				break;
#endif				
			default:
//				printf("Received: %02X\n", buffer[bytes_read]);
				bytes_read += res;
		};
	} while (!stop);
	
	buffer[bytes_read]=0;             /* set end of string, so we have a real string */

	return (bytes_read);
};


/* Convert a string to ISO 8859-15 
	Attention: Length of the string may change !
*/
void
utf8_to_iso8859_15(unsigned char *s){
	unsigned char *rd = s;
	unsigned char *wr = s;
	unsigned char b0, b1;
	unsigned int o1;
	
	while ( (b0 = *rd++) != '\0'){
		if (b0 & 0x80) {
			b1 = *rd;			// we may read next character here, because we have not hit '\0'.
			if ( ((b0 & 0xE0) != 0xC0) || ((b1 & 0xC0) != 0x80) ){
				/* error: not in the range of ISO 8859-15 */
				*wr++ = '?';
				continue;
			};
			/* We now know b1 is not 0. */
			rd++;
			
			o1 = ((b0 & 0x1f) << 6 ) | (b1 & 0x3f);
			if (o1 > 256) o1 = '?';
			*wr++ = o1;
		} else 
			*wr++ = b0;
	};
	*wr = '\0';
};

	

/* Read the response from mpd host
	Read one complete line
	Times out after 2.5 seconds
	Returns length of line (incl. '\n') if successful,
	else 0.
	Terminates the complete line with a 0.
*/
int read_response (int read_fd, char *buffer){
	int status;
	int done = 0;
	int bytes_read = 0;
	int n;
	
	do {
		if (bytes_read >= (BUFFER_SIZE - 1)) {
			fprintf(stderr, "Error, too many characters in a line!\n");
			return 0;		// Error, too many characters in a line!
		};
		
		status = check_read(read_fd);
		if (status == 0){
			fprintf(stderr, "Error, time out!\n");
			return 0;		// time out!
		};
	
		/* Read just one character until EOL */
		n =  read(read_fd, buffer+bytes_read, 1);
		if (n == -1) {
			fprintf(stderr, "Error when reading!\n");
			return 0;		// error!
		};
		if (n == 0) {
			fprintf(stderr, "EOF while reading host!\n");
			return 0;		// connection lost ?
		};
		done = (buffer[bytes_read] == '\n');
		bytes_read += n;

	} while (!done);
	buffer[bytes_read] = 0;
	
	/* Convert to ISO 8859-15 */
	utf8_to_iso8859_15((unsigned char *) buffer);
	
	return strlen(buffer);
};

void reset_transmitter(int serial_fd){
	char *s;
	char buffer[BUFFER_SIZE];
	int len;
	
	printf("Sending break\n");

	/* Send a break to reset device into ISP mode */
	tcsendbreak(serial_fd,3);
	
	/* Send an uppercase U to negotiate baud rate */
	buffer[0] = 'U';
	buffer[1] = 0;
	
	len = 0;
	/* Send U to serial line */
	if (write_all(serial_fd, buffer, 1) == -1) {
		perror("sendall");
		printf("We only sent %d bytes because of the error!\n", len);
	};		
	
	len = read_boot_response(serial_fd, buffer);
	buffer[len] = 0;
		
	/* Send U to serial line to check that we are in sync */
	buffer[0] = 'U';
	if (write_all(serial_fd, buffer, 1) == -1) {
		perror("sendall");
		printf("We only sent %d bytes because of the error!\n", len);
	};		
	
	len = read_boot_response(serial_fd, buffer);
	buffer[len] = 0;

	if (buffer[0] == 'U')
		printf("Baud rate successfully negotiated\n");
	else {
		printf("Error: Could not negotiate baud rate!\n");
		exit(20);
	};
	
	
	s=":0100000310EC";
	printf("Manuf. ID \t= ");
	if (0 != send_string(serial_fd,s,buffer) ) {
		exit(20);	;
	};
	printf("%s\n",buffer);
	if (0 != strncmp(buffer, "15", 2)) {
		fprintf(stderr, "ERROR: Device not recognized\n");
		exit(20);
	};
		
	/* Reset the device */
	printf("\nResetting device\n");
	s=":00000008F8";
	write_all(serial_fd, s, strlen(s));
	read(serial_fd, buffer, 13);		/* Read back the echo and forget it */
	tcflush(serial_fd, TCIOFLUSH);	
};

/* Start of program. */
double t0;

void init_time(){
	struct timeval start_time;
	gettimeofday(&start_time, NULL);
	t0 = 	(double) start_time.tv_sec +  ((double) start_time.tv_usec) / 1000000;
};


/* Print time since start of program to stdout */
void prt_time(){
	struct timeval t1;
	double diff;
	
	gettimeofday(&t1, NULL);
	diff = ((double) t1.tv_sec) +  ((double) t1.tv_usec) / 1000000 - t0;

	printf ("[%.1lf ] ", diff);
};



int main(int argc, char *argv[])
{
	int len, total_len, cmd_finished, stop;
	int clientSocket, serial_fd, 
	remotePort,
	status = 0;
	struct hostent *hostPtr = NULL;
	struct sockaddr_in serverName = { 0 };
	char buffer[BUFFER_SIZE];
	char buffer2[BUFFER_SIZE];
	char *remoteHost = NULL;
	char *serial_device;
	struct termios oldtio,newtio;
	
	int filter_lsinfo;
	
	cur_tx_len = 0;
	
	if (4 != argc)
 	{
		fprintf(stderr, "Usage: %s <serial_device> <serverHost> <serverPort>\n", argv[0]);
		exit(1);
	};
	
	serial_device = argv[1];
	remoteHost = argv[2];
	remotePort = atoi(argv[3]);

	/* Initialize time keeping */
	init_time();
	
	
      /* 
	Open modem device for reading and writing and not as controlling tty
	because we don't want to get killed if linenoise sends CTRL-C.
       */	
	serial_fd = open(serial_device, O_RDWR | O_NOCTTY ); 
	if (serial_fd <0) {perror(serial_device); exit(-1); }
	
	tcgetattr(serial_fd,&oldtio); /* save current port settings */

	bzero(&newtio, sizeof(newtio));
       /* 
	BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
	CRTSCTS : output hardware flow control (only used if the cable has
		all necessary lines. See sect. 7 of Serial-HOWTO)
	CS8     : 8n1 (8bit,no parity,1 stopbit)
	CLOCAL  : local connection, no modem contol
	CREAD   : enable receiving characters
	*/
	newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;

	/*
	IGNPAR  : ignore bytes with parity errors
	ICRNL   : map CR to NL (otherwise a CR input on the other computer
	will not terminate input)
	IXOFF    : enable XON/XOFF flow control on input
	otherwise make device raw (no other input processing)
	*/
	newtio.c_iflag = IGNPAR | IXOFF | IXON;

	/*
	Raw output.
	*/	
	newtio.c_oflag = 0;

	/* set input mode (non-canonical, no echo,...) 
	disable all echo functionality, and don't send signals to calling program
	*/
	newtio.c_lflag = 0;
 
	newtio.c_cc[VTIME]    = 90;   /* time out after 9 seconds */
	newtio.c_cc[VMIN]     = 0;   /* non-blocking read */
	
	/* now clean the modem line and activate the settings for the port */
	tcflush(serial_fd, TCIFLUSH);
	tcsetattr(serial_fd,TCSANOW,&newtio);
 
	/* Reset the transmitting device so we have a defined state */
//	reset_transmitter(serial_fd);

	/*
	 * need to resolve the remote server name or
	 * IP address 
	*/
	hostPtr = gethostbyname(remoteHost);
	if (NULL == hostPtr) {
		hostPtr = gethostbyaddr(remoteHost,
				 strlen(remoteHost), AF_INET);
		if (NULL == hostPtr) {
			perror("Error resolving server address");
			exit(1);
	 	}
	}

	serverName.sin_family = AF_INET;
	serverName.sin_port = htons(remotePort);
	(void) memcpy(&serverName.sin_addr, hostPtr->h_addr, hostPtr->h_length);


	
	clientSocket = -1;		// Socket not connected
	
	/*
	 * Client application specific code goes here
	 *
	 * e.g. receive messages from server, respond,
	 * etc. 
	*/	
	stop = 0;
	do {	
		printf("Getting cmd from serial line\n");
		cmd_finished = 0;
		filter_lsinfo = 0;		
			/* Read a command from serial line 
				and send it to mpd
				until command or command list is finished
			
				TODO here we should check for EOT character
					and not "command_list_end" or end of line
					change the check of EOT in read_cmd
			*/
		do {
			len = read_cmd(serial_fd, buffer);
			
			if (len == 0) continue;	
			prt_time();
			
			printf("BETTY:%s", buffer);
			
			// NOTE some ugly hack. 
			// The lsinfo command gets too much information from mpd.
			// Betty only wants to see lines beginning with "playlist:"
			// So we set a flag to filter unwanted lines from transmission over serial link
			if (0 == strncmp(buffer, "lsinfo\n", 7))
				filter_lsinfo = 1;
			
#if 0			
			cmd_finished = 1;
			if (strncmp(buffer, "command_list_begin", 18) == 0)
				cmd_finished = 0;
		
			if (strncmp(buffer, "command_list_end", 16) == 0)
				cmd_finished = 1;
#endif	
//			cmd_finished == (buffer[0] == EOT);
			cmd_finished = 1;		
			/* send the command line to mpd */
			// make sure we are connected
			while (-1 == clientSocket){
			
				clientSocket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
				if (-1 == clientSocket)
				{
					perror("socket()");
					continue;
				};
	
				status = connect(clientSocket, (struct sockaddr*) &serverName, sizeof(serverName));
				if (-1 == status){
					perror("connect() to mpd failed");
					continue;
				}
 
				// The mpd server responds to a connection with a version line beginning with "OK"
				if (!read_response(clientSocket, buffer2) ||strncmp(buffer2, "OK", 2))  {
					perror("bad initial response from mpd");
					continue;
				};
			};
				
			if (write_all(clientSocket, buffer, len) == -1) {
				perror("sendall");
				printf("We only sent %d bytes because of the error!\n", len);
			};		
		
		} while (!cmd_finished);

		total_len = 0;
		// get the response from mpd and send it to serial line
		printf("Getting response from MPD\n");
		do { 
			/* Read a complete line from mpd */
			len =  read_response(clientSocket, buffer);
			if (len == 0){			// There was an error
				stop = 0;
				break;
			};
			
			printf("MPD: %s", buffer);		
				
			if ( (filter_lsinfo == 0) || (strncmp(buffer, "playlist:", 9) == 0) ||
						   (0 == strncmp(buffer, "OK", 2)) || (0 == strncmp(buffer, "ACK", 3)) ){
				/* Send response to serial line */
				if ( (len = write_all_ack(serial_fd, buffer, len)) == -1) {
					perror("sendall");
				} else		
					total_len += len;
			
				if (0 == strncmp(buffer, "OK", 2))
					break;	
				if (0 == strncmp(buffer, "ACK", 3))
					break;	
			};
		} while (1);	
		
		buffer[0] = EOT;
		write_all_ack(serial_fd, buffer, 1);	// TODO check that this write really succeeded
		total_len += 1;
			prt_time();
			printf("MPD: Sent %d bytes to serial\n", total_len);
		close(clientSocket);
		clientSocket = -1;
	} while (!stop);
		

	/* restore the old port settings */
	tcsetattr(serial_fd,TCSANOW,&oldtio);

	return 0;
}
